#ifndef MOD_CONTROL_H
#define MOD_CONTROL_H

#include "PT/PTM_PartContLC.h"
#include "PT/PTM_ParPartCont.h"
#include "PT/PTC_Para.h"
#include "PT/PTBinaryIO.h"
#include "PT/PTstring.h"
#include "PT/PTS_Factory.h"
#include "PT/PTS_CommBase.h"
#include "PT/PTS_Unique.h"
#include "PT/PTfstream.h"

#include "ModParticle.H"   // DIM

typedef PTM_ParPartCont<MDRotParticle,DIM,PTM_PartContLC<MDRotParticle,DIM,double_vec,int_vec> > ParticleBox;

class Control;


// --------------- add your event declarations here


/** each keyword in the control file corresponds to one event. The
    data after the keyword is parsed by the constructor of the event
    which should therefore accept an istream argument. For each event,
    you must invent a class name, derive from 'Event' write the
    constructor and an 'action' routine, which figures out what to do
    from the data in your Event structure and the argument passed to
    the action routine.
 
    The event and its keyword must be registered in the constructor of 
    the Control.

    If the events need to store data aross program invocations in the restart 
    file, you must implement save/restore as well

    I have removed box.update from all the action routines. 
*/
struct Event {
  /** @return 0 if the event should be removed from the list, >0 otherwise */
  virtual int action(Control &, double time, ParticleBox &box, const Material &)=0;
  virtual std::ostream &save(std::ostream &out) const { return out; }
  virtual std::istream &restore(std::istream &in) { return in; }
  virtual ~Event(){}

  PTS_Unique<Event> unique; // identifier required to match on-file events 
                            // to those in the event_queue
};

// ---------------------------------------------------------------

struct LoadBallFile : public Event {
  // insert particles from positions specified in a configuration file

  LoadBallFile( istream &); // initializes data

  int action(Control &, double t, ParticleBox &, const Material &);

private:
  std::string my_filename;
};

// ---------------------------------------------------------------

struct Load : public Event {
  // load particles in intervals from specified input positions

  Load( istream  & );

  int action(Control &, double time, ParticleBox &, const Material &);

  virtual std::ostream &save(std::ostream &out) const;
  virtual std::istream &restore(std::istream &in);

private:

  struct input_point {
    input_point()
      : n(0), type(0), delay(0.), curr(0.), active(false) {}
    double_vec location;
    unsigned int n;
    unsigned int type;
    double delay, curr;
    bool active;
  };

  friend istream& operator>>(istream&, input_point&);
  friend istream& operator>(istream&, input_point&);
  friend ostream& operator<(ostream&, const input_point&);
  friend ostream& operator<<(ostream&, const input_point&);

  double factor(){ return 0.25; } // jitter factor to avoid exact columns

  std::vector<input_point> inp_point;
  vector<unsigned int>     num_of_ball; // typewise ball number limits
  unsigned      types;       // in principle size() of of num_of_balls
  double        start, stop; // time limits of filling process
};

// ---------------------------------------------------------------------

struct Remove : public Event {
  
  Remove ( istream & );

  int action(Control &, double time, ParticleBox &, const Material &);

  virtual std::ostream &save(std::ostream &out) const;
  virtual std::istream &restore(std::istream &in);

  double_vec ll, ur;
  double     start, stop, itvl;
  double     next;
  unsigned   limit;  // max number of particles to be removed each time
};

// =====================================================================

class Control {
private:
  
  // helper needed in erase algorithm to evaluate the event action routine 
  // and return the info whether the event should be removed from the 
  // event queue or not.
  struct Activate {
    Activate( Control &a_control, 
	      double a_time, ParticleBox &a_box, const Material &a_mat)
    : control(a_control), currtime(a_time), box(a_box), mat(a_mat) 
    { } 

    bool operator()(Event *e){
      return 0 == e->action(control,currtime,box,mat);
    }

    Control        &control;
    double          currtime;
    ParticleBox    &box;
    const Material &mat;
  };

  vector<Event *>   event_queue;
  unsigned          initial_queue_size; 
  const Material   &material;

  // produces events from the ASCII characterization in the input file
  static PTS_ExtTagFactory<Event,std::string> our_factory;

  // for a unique particle count
  static unsigned long our_particle_count;
  unsigned long        m_node_id;

  // allows for 1024 processors (10 bits)
  static const unsigned      our_proc_bits = 10;

  // some of the event classes need access to this member, as a quick hack, 
  // I declare them 'friend' - rather than providing full blown access to 
  // everybody
  friend class Remove;
  PTS_CommEnv    &m_cenv;

public:
  // we need per CPU stuff for the unique_id function and 
  // for communication within the events. That's not quite clean, but you 
  // are invited to think of a more modular way to pass this info
  Control(const Material &a_mat, PTC_Para &pf, PTS_CommEnv &a_cenv);

  // ~Control(){}

  // attempt at creating a uniqe particle id. 
  // will work across program invocations only with correct 
  //  save/restore and if the number of processes does not change.
  unsigned long unique();

  bool read(string strFile);

  void processEvents(double currtime, ParticleBox &box);

  void printStatus(ostream &out=cout);

  std::istream & restore(std::istream &in);

  std::ostream & save(std::ostream &out) const;

  int active_events() const { return event_queue.size(); }
};

#endif

















