#include "ModControl.H"
#include "PT/PTS_ParStream.h"

#include <algorithm>
#include <limits>



// ----------------------------------------------------------------------


LoadBallFile::LoadBallFile(istream &in){
  in >> my_filename;
}


// input particles at start of simulation from an ASCII configuration file
// must execute on ALL processors because add_end() requires 
//  collective communication and must therefore be called everywhere 
int
LoadBallFile::action(Control &ctl, double t, ParticleBox &box, const Material &){
  // parse file for balls and insert them, afterwards remove event from list

  ifstream tFile(my_filename.c_str());

  // applied to position and velocity (scale) in the order shift, scale: 
  double     scale = 1.;
  double_vec shift(0.,0.,0.);


  if(! tFile ) {
    PT_WARNING( "couldnt open ball configuration file '"  
		<< my_filename << "'" );
  }
  else {
    // input ok

    if ( t > 0. ) { 
      PT_WARNING("ignoring ball configuration event: not at start"
		 " of simulation (t=" << t << ")" );
    }
    else {

      int  count=0;  // number of added balls
      
      string strLine;
      while (tFile) {
	getline(tFile,strLine);
	
	// Is there a comment in this line?
	size_t tPos=strLine.find_first_of('!');
	if (tPos!=string::npos){
	  // Remove comment
	  strLine=strLine.erase(tPos);
	}
	
	// Get command string
	string strWord=ParseString(strLine);
	strWord=LowerCase(strWord);
	if ( strWord.size() == 0 ) continue;
	
	if (strWord=="ball"){
	  // setup a new particle
	  MDRotParticle s;
	  
	  stringstream strin(strLine);
	  strin >> s;          // position, velocity and type
	  if ( strin.fail() ){ // oohps problem
	    PT_SEVERE( "ball input failed, ignoring particle: " << s );
	    continue;
	  }
	  
	  // scaling (shift first, then scale)
	  s.x += shift;
	  s.x *= scale;
	  s.v *= scale;
 
	  // only if the particle has a chance to be kept in the box, 
	  // we assign an id to it
	  if ( s.x >= box.ll() && s.x <= box.ur() )  
	    s.id = ctl.unique(); // particle id
	  
	  box.add(s);            // add it to box
	  ++count;
	} 
	else if ( strWord=="scale" ) {
	  
	  stringstream strin(strLine);
	  strin >> scale;      // scale factor for position and velocity
	  if ( strin.fail() ){ // oohps problem
	    PT_SEVERE( "scale input failed" );
	  }
	  PT_INFO( "scale: " << scale );
	}
	else if ( strWord=="shift" ) {

	  stringstream strin(strLine);
	  strin >> shift;      // shift for position and velocity
	  if ( strin.fail() ){ // oohps problem
	    PT_SEVERE( "shift input failed" );
	  }
	  PT_INFO( "shift by: " << shift );
	}
	else {
	  PT_WARNING( my_filename << ": unknown key word `" << strWord << "'");
 	}
	
      } 
      // no more input
      
      if (count > 0) {
	box.add_end();
	PT_INFO( "added " << count << " particles from " << my_filename );
      }
    
    }  // t == 0
  }    // ball configuration file ok
  return 0;
}


// ------------------------------------------------------------------------
// loading particles in intervals at specific 'input points'

Load::Load( istream  &in ){

  // load first expects 'global' parameters:
  // start time, end time, then number of balls of each type
  // in each subsequent line containing the word 'input_point' a separate 
  // position, interval, type  specificiation is expected
  // an empty line or an end on a line ends the input of 'load'

  // first word was "load"
  PT_DEBUGINFO( "in load ctor" );

  in >> start >> stop >> types;
  for ( unsigned i=0; i<types; ++i ){
    unsigned tmp;
    in >> tmp;
    num_of_ball.push_back(tmp);
  }
  // ignore newline - load must be terminated, so there is either a new 
  // empty line or a line containing end_load
  while ( in.get() != '\n' );

  // now read input points
  while ( in ){
    string line, word;
    getline(in,line);

    string::size_type pos = line.find_first_of("!");
    if ( pos < PT_STR_NPOS ){ // found it
      line = line.erase(pos,PT_STR_NPOS);
      word = ParseString(line);
      word = LowerCase(word);
      if ( word == "" ) continue;
    }
    else {
      word = ParseString(line);
      word = LowerCase(word);
    }
    PT_DEBUGINFO( "key = " << word );
    
    if ( word == "input_point" ){ // ok
      input_point tmp;
      istringstream str_in(line);
      str_in >> tmp;
      inp_point.push_back(tmp);
    }
    else if ( word == "end_load" || word == "" ) {  // empty
      PT_DEBUGINFO( "load ctor: " << inp_point.size() << "input positions" );
      break;
    }
    else {
      PT_WARNING( "end_load or empty line missing at end of load section" );
      break;
    }
  }
}   // Load::Load 


// load particles in intervals from specified input positions
// must execute on ALL processors because add_end() requires 
//  collective communication and must therefore be called everywhere 
int 
Load::action(Control &ctl, double currtime, ParticleBox &box, const Material &material){
  
  int remaining_balls = 0;
  int count = 0;

  if ( currtime < start ) return 1; // keep for later
  if ( currtime > stop ) {
    PT_DETAILLED( "load event - stop time remove hint at t=" << currtime ); 
    return 0; // nevermore... 
  }

  for ( vector<input_point>::iterator i= inp_point.begin();
	i != inp_point.end();
	++i ){

    remaining_balls += num_of_ball[(*i).type];

    if (currtime >= (*i).curr && num_of_ball[(*i).type] > 0 ){

      PT_DEBUGINFO3("load action: pos, type, delay: " <<
		    (*i).location << " " << (*i).type << " " << (*i).delay );
      PT_DEBUGINFO3(" curr / num_of_ball " << (*i).curr << " " 
		    <<  num_of_ball[(*i).type] );

      MDRotParticle p;
		  
      p.x    = (*i).location;
      p.v    = double_vec(0.,0.,0.);
      p.type = (*i).type;

      const double r= material.getRadius(p.type);
      p.x += this->factor() * double_vec(r*rand()/RAND_MAX,
					 r*rand()/RAND_MAX,
					 r*rand()/RAND_MAX );
      
      // only if the particle has a chance to be kept in the box, 
      // we assign an id to it
      if ( p.x >= box.ll() && p.x <= box.ur() )  
	p.id = ctl.unique();

      box.add(p);
      count++;
      num_of_ball[(*i).type]--;

      (*i).curr = currtime + (*i).delay; // next insertion
    } 
  }  // for all tubes

  if ( remaining_balls == 0 ) {
    PT_DETAILLED( "load event - balls exhausted remove hint at " << currtime ); 
    return 0; //remove event from list
  }
  else if ( count > 0 )
    box.add_end();

  return 1;
}

std::ostream &
Load::save(std::ostream &out) const {
  PT_DEBUGINFO(" saving load ev. #" << unique.id() );
  return out < inp_point < num_of_ball < types; 
}

std::istream &
Load::restore(std::istream &in) {
  return in > inp_point > num_of_ball > types;
}


//   input operator for input_point
istream &
operator>>(istream &in, Load::input_point &l){
  return in >> l.location >> l.type >> l.delay;
} 

// binary i/o for input_point
std::ostream &
operator<<(std::ostream &out, const Load::input_point &ip){
  return out << ip.location << " " << ip.type << " " << ip.delay;
}

// binary i/o for input_point
std::ostream &
operator<(std::ostream &out, const Load::input_point &ip){
  return out < ip.location <  ip.n < ip.type < ip.delay < ip.curr < ip.active;
}

std::istream &
operator>(std::istream &in, Load::input_point &ip){
  return in > ip.location >  ip.n > ip.type > ip.delay > ip.curr > ip.active;
}


// -----------------------------------------------------------------

Remove::Remove( istream &in ){
  // just got "remove", expect vector vector start end itvl
  in >> ll >> ur >> start >> stop >> itvl;
  int a_limit;
  in >> a_limit; 
  if ( in.fail() ) { // trouble, probably limit was not present, use 1 
    PT_WARNING ( "no limit specfication for Remove, using 1" );
    a_limit = 1;
  }
  if ( a_limit < 0 || a_limit > 1000 ) { // probably another problem
    PT_WARNING ( "limit negative or very large: " << a_limit << "using 1" );
    a_limit = 1;
  }
  limit = static_cast<unsigned>(a_limit);
  next  = start;
} 

int 
Remove::action(Control &ctl, double time, ParticleBox &box, const Material &){

  // if it is our turn, we remove at most 'limit' (member) 
  // particle(s) from the quadrilateral specified by ll and ur;
  // for continous operation, this routine should be complemented
  // by a mechnism by which the balls are re-inserted at the top
  // (in the spirit of load). For now I think it is sufficient to 
  // have both a 'load' event and a 'remove' event and to match the rates 
  // at which particles are inserted/drawn

  if ( time > stop ) {
    PT_DETAILLED( "remove event - stop time remove hint" ); 
    return 0; // remove event
  }

  const int sort_dir = 2;   // sort according to z coordinate (x=0, y=1)

  // on each node, we put the particles inside the remove volume into
  // a vector, we then collect the stuff on CPU 0 and sort according to
  // height. Then we decide which particles to remove and communicate
  // that information back to the participating CPUs 

  if ( time > next ){ // next is at least equal to the start time at 
                      // the beginning

    typedef std::pair<double,unsigned long> PartInfo;

    std::vector< PartInfo > candidates;

    for ( ParticleBox::iterator i=box.begin(); i!=box.end(); ++i ){
      // compare componentwise with AND (provided by PTZ_Vector)
      if ( ll <= (*i).x && ur >= (*i).x ){
	candidates.push_back( PartInfo( (*i).x[sort_dir], (*i).id) );
      }
    }  // figured out the particle on this node in the removal vol.

    if ( ctl.m_cenv.node_id() == 0 ) { 

      // master node: copy our stuff into a multimap, then insert the stuff 
      // from all other CPUs in turn 
      // sort, make a vector of the results, broadcast that 
      // vector back; finally remove the particles with id's matching those 
      // in our simulation volume

      std::multimap<double,unsigned long> particles; 
      for ( std::vector< PartInfo >::iterator it = candidates.begin(); 
	    it != candidates.end();
	    ++it )
	particles.insert( *it );
      
      // now for each other CPU
      for ( int inode=1; inode < ctl.m_cenv.n_nodes(); ++inode ){
	PTS_ParStream   channel(ctl.m_cenv,inode);
	channel > candidates;   // read candidates
	
	for ( std::vector< PartInfo >::iterator it = candidates.begin(); 
	      it != candidates.end();
	      ++it )
	  particles.insert( *it );
      }
      
      // now we put the particles matching the removal criterion 
      // into the candidates vector and communicate them back
      // multimap has sorted the particles according to what we gave as 
      // first argument of the pair
      candidates.clear();
      candidates.reserve(limit); // at most this much
      int count = 0;
      while ( particles.begin() != particles.end() 
	      && count < static_cast<int>(limit) ) {  // there are still some
	candidates.push_back( *(particles.begin()) );
	particles.erase ( particles.begin() );
	++count;
      }

      if ( candidates.size() < limit ) 
	PT_WARNING( "removing less particles than requested: " <<
		    candidates.size() << " (" << limit << ")" );

      // send the remove list to everybody
      ctl.m_cenv.broadcast( candidates, 0 );

    }
    else {

      // node_id != 0  -- just send our candidates to CPU 0

      PTS_ParStream   channel(ctl.m_cenv,0);
      channel < candidates;

      // read what CPU 0 says should be removed
      ctl.m_cenv.broadcast( candidates, 0 );

      if ( candidates.size() < limit ) 
	PT_WARNING( "removing less particles than requested: " <<
		    candidates.size() << " (" << limit << ")" );

    }

    // particle removal now takes place on all CPU's

    // check wether the particle id's match any in our box, 
    // if yes erases them. Please verify that this algorithm is ok 
    // with any other choice of parallel particle container than 
    // PTM_ParPartCont, the general idea is that we don't have to consider 
    // the 'low' positions of the two containers again, 
    // since they have had a chance already to comapre equal once we are 
    // beyond them. The erase (see comment) will have the effect to 
    // swap or copy particles from positions larger than the current one
    // so that we may not move the iterators to not miss particles that 
    // have taken the place of the erased ones. 

    ParticleBox::iterator it = box.begin();
    while ( it != box.end() ) {
      std::vector<PartInfo>::iterator jt = candidates.begin();
      bool box_unchanged = true;
      while ( jt != candidates.end() ) {
	if ( (*it).id == (*jt).second ) { 
	  // sts: I have no ideas what the standard guarantees for the 
	  // validity of the iterators after this operation,
	  // I checked that this will work with PTM_ParPartCont 
	  // and with the vector implementation in KAI's Modena library
	  box.erase( it );
	  candidates.erase( jt );
	  box_unchanged = false;
	  // do not move the iterators so that they have a chance to 
	  // compare equal to a possibly modified end()
	}
	else  
	  ++jt; // next candidate
      }
      if ( box_unchanged ) ++it;
    }
    
    // must be called on all proc's simultaneously
    box.add_end();
    next += itvl;
  }

  return 1; // leave event in queue
}

std::ostream &
Remove::save(std::ostream &out) const {
  PT_DEBUGINFO(" saving load ev. #" << unique.id() );
  return out < ll < ur < start < stop < itvl < next < limit;
}

std::istream &
Remove::restore(std::istream &in) {
  return in > ll > ur > start > stop > itvl > next > limit;
}


// -------------------  CONTROL  -----------------------------

// static variables
PTS_ExtTagFactory<Event,std::string> Control::our_factory;
unsigned long int                    Control::our_particle_count;


Control::Control(const Material &a_mat, PTC_Para &pf, PTS_CommEnv &a_cenv) 
  : material(a_mat), m_node_id(a_cenv.node_id()), m_cenv(a_cenv)
{
  // change our_proc_bits if you need more than 1024 processors
  // the processor id is part of the particle id in order to 
  // decouple particle number assignment on different processors
  if ( m_node_id >= (1u<<our_proc_bits) )
    PT_SEVERE ( "only " << (1u<<our_proc_bits) << " processors supported");

  static int invocations = 0;
  if ( invocations == 0 ){
    // register possible event types with allowed key strings
    our_factory.reg<Remove>("remove");
    our_factory.reg<Load>("load");
    our_factory.reg<LoadBallFile>("loadballfile");
    invocations++;
  }

  string strControl = pf.getd("project_name", ""); //name of control file
  PT_ASSERT( strControl != "" );
  strControl += ".ctrl";
  ifstream in(strControl.c_str());
  if ( !in ){
    PT_SEVERE( "Cannot open control file '" << strControl << "'" );
    exit(-1);
  }

  std::string line;
  do { 
    getline(in,line);
    std::string::size_type pos = line.find_first_not_of(" \t\n");
    if ( pos == std::string::npos ) 
      continue; 
    else if ( line[pos] == '!' ) // comment
      continue;
    else {

      PT_DEBUGINFO( line );
      istringstream str_in(line);
      std::string key;
      str_in >> key; 
     
      PT_DEBUGINFO( key );
      PT_ASSERT( key == "remove" || key == "load" || key == "loadballfile" );
      if ( key == "load" ){
	// load must be alone on a line, input may be multiline
	event_queue.push_back( our_factory.make(key,in) );
      }
      else
	// single line input from str_in
	event_queue.push_back( our_factory.make(key,str_in) );
    }
  } while ( in.good() );
  
  initial_queue_size = event_queue.size();
  PT_INFO( "found " << event_queue.size()  
	            << " events in configuration file " << strControl ); 
}


void
Control::processEvents(double currtime, ParticleBox &box)
{
  // runs a loop over all elements, thereby removing the events that 
  // return 0

  Activate activate(*this,currtime,box,material);
  event_queue.erase( remove_if(event_queue.begin(),event_queue.end(),
			       activate ),  // gives new end
		     event_queue.end()      // old end 
		   );

}


void 
Control::printStatus(ostream &out) 
{
  out << "registered " << event_queue.size() << " events" << std::endl;
}


std::istream &
Control::restore(std::istream &in) {

  in > our_particle_count;

  // events from the stored initial_queue_size up are new, we
  // dont touch them; all others, which we do not find in the restart file
  // will be removed from the event queue
  in > initial_queue_size;  
  PT_ASSERT( event_queue.size() >= initial_queue_size );

  // mark still active events
  vector<bool>              active(event_queue.size(),false);
  for ( unsigned i=initial_queue_size; i < event_queue.size(); ++i )
    active[i] = true;

  vector<Event *>::size_type size;

  in > size;
  PT_DEBUGINFO4("found size " << size );
  unsigned old_event=0;
  for( vector<Event *>::size_type i=0; i < size; ++i ){
    unsigned id;
    in > id;
    PT_DEBUGINFO4("found id " << id );

    // increase the event pointer until we find 'our' event
    while (    id != event_queue[old_event]->unique.id() 
	    && old_event < initial_queue_size ) 
      ++old_event;

    // if id's are missing, there were events removed from the .ctrl file, 
    // you may only ADD in the course of one simulation
    // since we do not know the type of the original event, we are left 
    // with an inconsistent state of the restart file now.
    PT_ASSERT( old_event != initial_queue_size );

    // otherwise, we restore the data for the current event
    in > *event_queue[old_event];   // calls restore virtually
    active[old_event] = true;
  }
  // keep the current queue size in mind fro the next restart
  initial_queue_size = event_queue.size();

  // copy/keep all active events
  vector<Event *>  new_queue;
  for ( unsigned i=0; i < event_queue.size(); ++i )
    if ( active[i] ) new_queue.push_back( event_queue[i] );
  event_queue.swap( new_queue );

  return in;
}

std::ostream &
Control::save(std::ostream &out) const {
  out < our_particle_count;
  out < initial_queue_size; // needed because events might have been 
                            // removed up to date
  out < event_queue.size();
  for ( vector<Event *>::size_type i=0; i < event_queue.size(); ++i ) {
    out < event_queue[i]->unique.id();
    out < *event_queue[i];
  }
  return out;
}

unsigned long 
Control::unique(){
  unsigned long result = our_particle_count++;
  //const unsigned long proc_mask = 
  //   ~0ul << (std::numeric_limits<unsigned long>::digits - our_proc_bits); 

  PT_ASSERT( sizeof(unsigned long) >= 4 );
  const unsigned long proc_mask = 
     ~0ul << (32 - our_proc_bits); 
  PT_DEBUGINFO( "proc mask " << proc_mask );

  PT_SOMECHECKS(
  if ( (result & proc_mask) != 0ul ) // ok, we dont loose high value bits
    PT_WARNING( "particle count overflow on node " << m_node_id << 
		" at " << our_particle_count << 
		"(" << (result & proc_mask) << ")" );
  )

  // replace high value bits by processor id
  //  result |= m_node_id << 
  //          (std::numeric_limits<unsigned long>::digits - our_proc_bits);
  result |= m_node_id << (32 - our_proc_bits);
  return result;
}












