#ifndef MODFORCE_H
#define MODFORCE_H

#include "ModParticle.H"
#include "ModWall.H"
#include "PT/PTmap.h"  // needed by contact verification

// -------------------------------------------------
//  particle-particle force 
//  particle-wall     force
//  contact verification/bogus elimination 
// -------------------------------------------------

#ifndef  MOD_TANGENTIAL_DAMPING 
# define MOD_TANGENTIAL_DAMPING 1
#endif

class ForceCompute{
public:
  ForceCompute(const Material &a_mat) : material(a_mat) { }


  template<class P> inline  
  double plastic_force(P & p1, P & p2, double overlap);
  template<class P> inline  
  double force_loading_Hertz( P & p1, P & p2, double overlap );
  template<class P> inline  
  double force_unloading_Hertz( P & p1, P & p2, double overlap,double xmax); 
  template<class P> inline  
  double compute_xp( P & p1, P & p2,double xmax);
  template<class P> inline  
  double decision( P & p1, P & p2, double overlap,double xmax);
//  template<class P>
//  double WallFeature::compute_xp(double xmax);
private:
  const Material material;	
};



class ForceModel1 {
public:
  // handles collision of two spheres
  template<class P> inline void 
  operator()( P & p1, P & p2 );
  
  ForceModel1(const Material &a_mat)
  : material(a_mat), compute(a_mat) {}

private:
  const Material material;
  ForceCompute compute;
};

// =================  contact model  ====================================
// dont forget to update the wall contact model if you make changes here!

template<class P> 
void 
ForceModel1::operator()( P & p1, P & p2 ){
 
  double_vec diff;                      // difference between centers of
                                        // spheres
  double sum_r =   material.getRadius(p1.type)
                 + material.getRadius(p2.type);
  // calculate difference and distance^2
  diff = p1.x - p2.x;   // vector pointing from second to first particle 
  const double dist2 = dot(diff,diff);

  // was there a collision??
  if ( dist2 <= (sum_r*sum_r) ){
    
    const double dist = sqrt(dist2);  // distance  
    const double h    = sum_r-dist;   // always positive or zero 

    PT_DEBUGINFO3( "particle " << p1.id << " x/v/omega: " << p1.x << p1.v << p1.omega );
    PT_DEBUGINFO3( "particle " << p2.id << " x/v/omega: " << p2.x << p2.v << p2.omega );


#if MOD_ROTATION

    const double frac1 = material.getRadius(p1.type) / sum_r;
    const double_vec lever1  = ( -frac1) * diff;  // center1  -> contact 
    const double_vec lever2  = (1-frac1) * diff;  // center2  -> contact

    // surface realtive velocity seen from part. 1
    // omega_i x r_i are the surface vel. due to rotation. We must count the 
    // second positive and the first negative if we want to know the 
    // relative velocity as seen from the first particle (check: 
    // the contributions cancel for opposed omega_i) 
    double_vec vrel   = p2.v - p1.v - cross(p1.omega,lever1)
                                    + cross(p2.omega,lever2); 

#else
    double_vec vrel   = p2.v - p1.v;      // rel particle velocity
#endif

    // Hertzian contact model
    const double sqrt_h = sqrt(h);
    
    const double damp  = material.getDamp(p1.type,p2.type)      * sqrt_h;
  //  compute force move after MOD_PERSISTENT_CONTACTS
  //  const double force = material.getForce(p1.type,p2.type) * h * sqrt_h;//zmiana wywolania sily==
#if MOD_TANGENTIAL_DAMPING
    const double tdamp = material.getTDamp(p1.type,p2.type) * sqrt(sqrt_h);
#endif
#if MOD_JANSEN_PRESSURE
    const double force1 = material.getForce(p1.type,p2.type) * h * sqrt_h;//zmiana wywolania sily==
    p1.pressure += force1;
    p2.pressure += force1;
#endif

    PT_ASSERT( force >= 0. );

    double_vec normal = (1./dist) * diff; // unit vector from 2nd to 1st
    double_vec v_norm = dot(vrel,normal) * normal;
    double_vec v_tang = vrel - v_norm;    // orient. of normal doesn't matter
    
#if MOD_PERSISTENT_CONTACTS
    PT_DEBUGINFO3( "particle " << p1.id << ", no of contacts=" << p1.contact.size() ); 

    // if there was a collision, I need to check whether the coll partner 
    // is in the coll.partner list and if necessary insert it.

    typename P::CMapIterator  it1, it2;
    it1 = p1.contact.find( p2.id );

    if ( it1 == p1.contact.end() ) {

      PT_SAFE( 
      if ( p1.contact.size() >= p1.contact.capacity() ) {
	PT_WARNING("particle " << p1.id << ": more than " 
		   << p1.contact.capacity() << " neighbors (" 
                   << p1.contact.size() << "): " ); 
	for ( typename P::CMapIterator it3 = p1.contact.begin();
              it3 != p1.contact.end();
              ++it3 ) {
           PT_CERR << (*it3).first << " ";
        }
      }
      )  // safe

      // insert a new contact, use v_tang as relative velocity 
      it1 = p1.contact.insert(std::pair<unsigned long,MDContact>( p2.id,
				       MDContact(double_vec(0.), v_tang)) );

      PT_ASSERT( p2.contact.find( p1.id ) == p2.contact.end() ); 
      PT_SAFE( 
      if ( p2.contact.size() >= p2.contact.capacity() ) {
	PT_WARNING("particle " << p2.id << ": more than " 
		   << p2.contact.capacity() << " neighbors (" 
                   << p2.contact.size() << "): " ); 
	for ( typename P::CMapIterator it3 = p2.contact.begin();
              it3 != p2.contact.end();
              ++it3 ) {
           PT_CERR << (*it3).first << " ";
        }
      }
      )  // safe

      it2 = p2.contact.insert(std::pair<unsigned long,MDContact>( p1.id,
				       MDContact(double_vec(0.), -v_tang)) );
    }
    else {
      it2 = p2.contact.find( p1.id );
      // mark old contacts as active
      (*it1).second.active = true;
      (*it2).second.active = true;

    }

    double_vec & delta_s1     = it1->second.delta_s;
    double_vec & delta_s1_dot = it1->second.delta_s_dot;
    double_vec & delta_s2     = it2->second.delta_s;
    double_vec & delta_s2_dot = it2->second.delta_s_dot;

    // this integration can be done here because we will consider 
    // each contact exactly once (in p-w-interaction and p-p-interaction)
    //  predictor step:
    delta_s1         += delta_s1_dot * material.get_timestep_gpk5(); 
    // project into tangent space before computing the force
    delta_s1         -= dot(delta_s1,normal) * normal;
    // friction points _in_ the direction of delta_s which is the perceived
    //  surface velocity of the second particle
    double_vec fric  = material.getFric(p1.type,p2.type) * delta_s1 * sqrt_h;
   // override force but in pressure particle is still old force !!IMPROVE!! 
    const double force = compute.decision(p1,p2,h,(*it1).second.overlap_max);//zmiana wywolania sily==
     // decide which overlap is greater
     if((*it1).second.overlap_max  < h)
	{
	 (*it1).second.overlap_max = h;
      	 (*it2).second.overlap_max = h;
	}
     //std::cout<<"typ1 "<<(int)p1.type<<" typ2 "<<(int)p2.type<<std::endl;	
     //std::cout<<material.getMass((int)p1.type)<<" "<<material.getMass((int)p2.type)<<std::endl;	
     //std::cout<<"KULKI zpx1: "<<p1.x[2]<<" zpx2: "<<p2.x[2]<<" force: "<<force<<" zpv1: "<<p1.v[2]<<" zpv2: "<<p2.v[2]<<" overlap: "<<h<<std::endl;  	
     PT_DEBUGINFO3( "  current delta_s1 " << delta_s1 ); 
#else
    // dummies for the case that we don't treat static friction
    double_vec delta_s1, delta_s2, delta_s1_dot, delta_s2_dot;
    double_vec fric = double_vec(0.);

//    const double force = material.getForce(p1.type,p2.type) * h * sqrt_h;//zmiana wywolania sily==
#endif

    // if static friction is too large to be supported by the contact, 
    //  switch to sliding friction 
    double stat_fric2 = dot(fric, fric);
    double dyn_fric   = material.get_mue(p1.type,p2.type) * force;
    
    if ( stat_fric2 > dyn_fric * dyn_fric ) {
      
      PT_DEBUGINFO3( "sliding contact " << p1.id << "<->" << p2.id );

      // direction delta_s, magnitude dyn_fric:
      // see comments in ModWall.H
      fric = dyn_fric * fric / sqrt( stat_fric2 );
      delta_s1  = fric / (sqrt_h * material.getFric(p1.type,p2.type));
    }
    else {
      delta_s1 += (v_tang - delta_s1_dot) * material.get_timestep_gpk5();
      
      // in the static case, we need to introduce a velocity proportional 
      // friction term to damp the tangential eigenmodes
#if MOD_TANGENTIAL_DAMPING
      fric += tdamp * v_tang;
      PT_DEBUGINFO3("  (p-p, static) tang. damping f: " << tdamp * v_tang); 
#endif
    }
    delta_s1_dot =  v_tang;
    delta_s2     = -delta_s1;
    delta_s2_dot = -delta_s1_dot;

    PT_DEBUGINFO3( "  final s/d. friction/v_tang " << fric << " " << v_tang);

    p1.f += (force*normal + damp*v_norm + fric);
    //  use actio = reactio (we iterate only over PAIRS):
    p2.f -= (force*normal + damp*v_norm + fric);  

#if MOD_ROTATION
    PT_DEBUGINFO3( "  final torques p1=" <<  cross(lever1,fric) <<
		   ", p2="               << -cross(lever2,fric) );

    // fric is frictional tangential force on particle 1, thus 
    // the torque is the expected  r x F = lever1 x fric 
    p1.torque  = cross( lever1, fric );
    // seen from particle 2, the friction changes sign
    p2.torque -= cross( lever2, fric );

#endif
  }
#if MOD_PERSISTENT_CONTACTS
  else { // no contact, make sure to remove the p1-p2 pair from contact lists

    typename P::CMapIterator it;
    it = p1.contact.find( p2.id ); 
    if ( it != p1.contact.end() ) {
      // contact existed before - it should not have been encountered as 
      // active before
      PT_ASSERT( (*it).second.active == false );
      p1.contact.erase( it );
      p2.contact.erase( p2.contact.find( p1.id ) );
     }

  }       // (else) removal from contact lists
#endif
}

//=================================================================
//=================================================================

/** ---------------------------------------------------------------
    Make decision in which state is particle contact. 
    Depend on which state is actually for contact we can decide how
    to compute normal force. 
    There are five state
    1 - elastic coming up to plastic
    2 - elastic coming down from state 1, not achieve plastic
    3 - plastic always coming up
    4 - elastic coming down from plastic
    5 - unloading from state 4
!! Must be include MOD_PERSISTENT_CONTACTS to invoke this function !!
!! TO DO improve MOD_PERSISTENT_CONTACTS module !!
    ---------------------------------------------------------------
  */
template<class P> 
double  
ForceCompute::decision( P & p1,  P & p2,double overlap,double xmax){

double normal_force = 0.;
double xy = material.getxy(p1.type,p2.type) ;
int g;

g = p1.stan;

if(overlap <= xmax)
	{
	if(xmax > xy)
		{
			double xp = compute_xp(p1,p2,xmax);
			if(overlap > xp)
			{
				//state 4
				normal_force = force_unloading_Hertz(p1,p2,overlap,xmax);
				if(p1.stan == 4)
					{
						p1.control++;
					}
				else
					{
						p1.control = 0;
					}
				if(p1.stan != 4)
				p1.stan = 4;

			}
			else
			{
				//state 5	
				normal_force = 0.0;

				if (p1.stan != 5)
				p1.stan = 5;
			}
		}	
	else
		{
			//state no. 2
			normal_force = force_loading_Hertz(p1,p2,overlap);

			if (p1.stan != 2)
			p1.stan = 2;
		}
	}
else
	{
		
		if(overlap > xy)
		{
			//state 3
			normal_force =plastic_force(p1,p2,overlap) ;

			if (p1.stan != 3)
			p1.stan = 3;
		}
		else
		{
			//state 1
			normal_force = force_loading_Hertz(p1,p2,overlap);

			if (p1.stan != 1)
			p1.stan = 1;
		}
	}
return normal_force;
};


//=================================================================

/** ---------------------------------------------------------------
    compute particle - particle force in plastic model
    ---------------------------------------------------------------
  */
template<class P> 
double ForceCompute::plastic_force(P & p1, P & p2, double overlap)
{
	double normal_force;
	normal_force =0.0 ;
	normal_force = material.getNForceYield(p1.type,p2.type) + M_PI*material.getpy(p1.type)*pow(material.getRR(p1.type,p2.type),2)*(overlap - material.getxy(p1.type,p2.type));
	return normal_force;
};


//=================================================================

/** ---------------------------------------------------------------
    compute particle - particle force in load Hertz model
    ---------------------------------------------------------------
  */

 template<class P> 
 double ForceCompute::force_loading_Hertz(P & p1, P & p2, double overlap)
{
	double normal_force; 
	normal_force = material.getForce(p1.type,p2.type) * overlap * sqrt(overlap);
	return normal_force;
};


//=================================================================

/** ---------------------------------------------------------------
    compute particle - particle force in unload Hertz model
    ---------------------------------------------------------------
  */
template<class P>
double ForceCompute::force_unloading_Hertz(P & p1,P &p2, double overlap,double xmax)
{
	double normal_force, xp, xe;
	xe = ( 3 * plastic_force(p1,p2,xmax))/ (4 * material.getEE(p1.type,p2.type) * material.getRR(p1.type,p2.type));
	//xe = xe * sqrt(xe);
	xe = exp(log(xe)/3) * exp(log(xe)/3);
	xp = xmax - xe;
	//normal_force = (4 * material.getEE(p1.type,this->get_type()) * sqrt(material.getRR(p1.type,this->get_type())) * (overlap - xp) * sqrt(overlap - xp)) / 3 ;
	normal_force = material.getForce(p1.type,p2.type) * (overlap  - xp) * sqrt(overlap - xp);

	return normal_force;
};

//=================================================================

/** ---------------------------------------------------------------
    compute xp variable
    ---------------------------------------------------------------
  */

template<class P>
double ForceCompute::compute_xp(P & p1,P &p2,double  xmax)
{
	double xp, xe;
	xe = ( 3 * plastic_force(p1,p2,xmax))/ (4 * material.getEE(p1.type,p2.type) * (material.getRR(p1.type,p2.type)));
	//xe = xe * sqrt(xe);
	xe = exp(log(xe)/3) * exp(log(xe)/3);
	xp = xmax - xe;
	return xp ;
};




//=================================================================

/** ---------------------------------------------------------------
    compute particle-wall interaction, solely based on distance and 
    normal from a virtual call to overlap 
    ---------------------------------------------------------------
  */
template<class P>
inline void
WallFeature::operator()(P &p){
  
  double roughness = 1; 
  double dist;
  double_vec normal;
  double_vec wall_vel = 0.;
  this->overlap(p.x, dist, normal, wall_vel, roughness);  // virtual call 
  
  const double r = material.getRadius(p.type);

  // was there a collision between this wall and a sphere?
  const double h = r - dist; 
  if ( h >= 0 ) {
    // yes, contact  
    PT_DEBUGINFO3( "contact with " << this->get_id() << 
		   " dist/normal/roughness " << dist << " " << normal << 
		                                        " " << roughness );
    PT_DEBUGINFO3( "particle " << p.id << " x/v/omega: " << p.x << p.v << p.omega );

    // split relative velocity into normal and tangential velocities
    double_vec v_rel  = p.v - wall_vel;  // as seen from the wall
    double_vec v_norm = dot(v_rel,normal) * normal;  // 'positive' for 
                                                   // receding particles
#if MOD_ROTATION   
    // v_tang velcoity of the wall seen from the particle:
    // contribution of the rotation to the relative velocity seen _from_ the 
    // particle is - (omega x r), with r = - dist . normal, since normal 
    // points from the contact to the center of the particle
    double_vec v_tang = - v_rel + v_norm + cross( p.omega, dist*normal );
#else 
    double_vec v_tang = - v_rel + v_norm;
#endif    

    // Hertzian contact model
    const double sqrt_h = sqrt(h);
    const double damp   = material.getDamp(p.type,this->get_type())
                              * sqrt_h;
#if MOD_TANGENTIAL_DAMPING
    const double tdamp  = material.getTDamp(p.type,this->get_type())
                              * sqrt( sqrt_h );
#endif
  //  const double force  = material.getForce(p.type,this->get_type())
 //                             * h * sqrt_h;
    PT_ASSERT( force >= 0. );

#if MOD_JANSEN_PRESSURE
   const double force1 = material.getForce(p.type,this->get_type()) * h * sqrt_h;//zmiana wywolania sily==
   p.pressure += force1;
#endif 

#if MOD_PERSISTENT_CONTACTS
    // if there is a part-wall contact, we must check whether the 
    // wall is in the contact list. If it is not, we must insert it.
    // At the end of this section, we should have a valid 'contact pointer'
    // to be updated in the static friction computation

    // as for map, operator[] will insert the element. if it does not yet 
    //  exist, then delta_s will get its value on construction (0,0,0) 
    //double_vec & delta_s     = p.contact[ this->get_id() ].delta_s;
    //double_vec & delta_s_dot = p.contact[ this->get_id() ].delta_s_dot;

    typename P::CMapIterator it;
    it = p.contact.find( this->get_id() );
    if ( it == p.contact.end() ) {
      // insert a new contact, use v_tang as relative velocity 

      PT_SAFE( 
      if ( p.contact.size() >= p.contact.capacity() ) {
	PT_WARNING("particle " << p.id << ": more than " 
		   << p.contact.capacity() << " neighbors (" 
                   << p.contact.size() << "): " ); 
	for ( typename P::CMapIterator it3 = p.contact.begin();
              it3 != p.contact.end();
              ++it3 ) {
           PT_CERR << (*it3).first << " ";
        }
      }
      )  // safe

      it = p.contact.insert(std::pair<unsigned long,MDContact>(this->get_id(),
				      MDContact(double_vec(0.), v_tang)) );//++
    }
    else {
      (*it).second.active = true; 
      // make sure we mark used contacts as active
    }

    double_vec & delta_s     = it->second.delta_s;
    double_vec & delta_s_dot = it->second.delta_s_dot;
    
    // we use predictor-corrector type 2nd order integration because 
    // a first order integrator introduces so much energy into the system 
    // that a particle sitting in a corner does not allow the friction 
    // forces to go to stationary values but rather introduces periodic 
    // cycles. For the moment I dont see how to avoid further data elements 
    // in the particle data structure for this purpose. 

    // this integration can be done here because we will consider 
    // each contact exaclty once (in p-w-interaction and p-p-interaction)
    //  predictor step
    delta_s          += delta_s_dot * material.get_timestep_gpk5(); 
    // project into tangent space before computing the force
    delta_s          -= dot(delta_s,normal) * normal;
    // as delta_s is in the direction of the relative velocity seen from 
    // the particle, the force experienced should be in the direction 
    // delta_s
    double_vec fric  = material.getFric(p.type,this->get_type()) 
                          * delta_s * sqrt_h;
  // override force but in pressure particle is still old force !!IMPROVE!! 
    const double force = decision(p,h);//zmiana wywolania sily==
     // decide which overlap is greater
 if(p.xmax  < h)
	{
	 p.xmax = h;
	}
	
      PT_DEBUGINFO3( "  current delta_s " << delta_s ); 
#else
    // dummies for the case that we don't actually treat static friction
    double_vec delta_s, delta_s_dot;
    double_vec fric = double_vec(0.);
#endif

    // if static friction is too large to be supported by the contact, 
    //  switch to sliding friction
    double stat_fric2 = dot(fric, fric);
    double dyn_fric   = material.get_mue(p.type,this->get_type()) 
                           * roughness * force;
    if ( stat_fric2 > dyn_fric * dyn_fric ) {

      // must use sliding friction value
      PT_DEBUGINFO3( "contact is sliding" );

      // create a vector with amplitude dyn_fric and direction delta_s
      //  this prescription deviates from what is normally taken as 
      //  direction (-v_tang) because the latter is arbitrarily badly
      //  defined when static friction is required to hold particles in 
      //  place. We rather use the direction accumulated frm the past motion 
      //  in delta_s (and use fric instead, because we have computed its
      //  squared length already)
      fric = dyn_fric * fric / sqrt( stat_fric2 );

      // reduce static friction as to become equal to sliding friction
      // ds -= v_tang * material.get_timestep_gpk5();

      // set delta_s equal to what would be required to create dyn_fric
      // motivation: if the motion changes direction, the following update 
      //  of delta_s should reduce delta_s such that again the static 
      //  case applies. The friction will then change sign if the 
      //  elastic energy store in the static-friction spring has been used
      //  up
      // it is wrong to reduce delta_s to 0 because then we will constanlty
      //  switch between static and sliding friction in the case where we 
      //  just slide with constant velocity. Sice we woud then do timesteps 
      //  with delta_s values too small, we'd effectively renormalize mu
      delta_s  = fric / (sqrt_h * material.getFric(p.type,this->get_type()));
    } 
    else {
      // update displacement in order to reflect relative motion 
      //  (static friction case)
      //  corrector step
      delta_s += (v_tang - delta_s_dot) * material.get_timestep_gpk5(); 

      // in the static case, we need to introduce a velocity proportional 
      // friction term to open dissipation channels
#if MOD_TANGENTIAL_DAMPING
      // we will never get here if we do not have persistent contacts
      fric    += tdamp * v_tang;
      PT_DEBUGINFO3("  (wall, static) tang. damping f: " << tdamp * v_tang); 
#endif
    }
    // update derivative
    delta_s_dot =  v_tang;

    PT_DEBUGINFO3( "  final s/d. friction/v_tang " << fric << " " << v_tang);
    PT_DEBUGINFO3( "  final force " << (force*normal - damp*v_norm + fric) );
   // std::cout<<h<<"="<<force<<"="<<v_norm[2]<<"\n";
    p.f      += (force*normal - damp*v_norm + fric);

    // collect force statistics
    this->sample( dot(force*normal-damp*v_norm,normal) );

#if MOD_ROTATION
    // only tangential components contribute to the torque,
    // we compute (fric x dist*normal) as (r x fric) to take the fact into 
    // acount that dist*normal points to the center of the particle instead 
    // pointing from the center to the contact point
    PT_DEBUGINFO3( "  final torque " << cross( fric, dist*normal) );

    p.torque += cross( fric, dist*normal ); 
#endif
  }
#if MOD_PERSISTENT_CONTACTS
  else {  
    // make sure that the wall does not remain in the particles contact list
    // once the contact has opened 
    p.xmax = 0.0;	
    p.stan = 0;
   // std::cout<<"koniec kontaktu---"<<p.f<<"---\n";
    //p.f = 0.0;
    typename P::CMapIterator  it;
    it = p.contact.find( this->get_id() );
    if ( it != p.contact.end() ) { 
      PT_ASSERT( ! (*it).second.active );
      p.contact.erase(it);
    }

  } // (else) removal from contact lists   
#endif

}
//=================================================================

/** ---------------------------------------------------------------
    Make decision in which state is particle contact. 
    Depend on which state is actually for contact we can decide how
    to compute normal force. 
    There are five state
    1 - elastic coming up to plastic
    2 - elastic coming down from state 1, not achieve plastic
    3 - plastic always coming up
    4 - elastic coming down from plastic
    5 - unloading from state 4
!! Must be include MOD_PERSISTENT_CONTACTS to invoke this function !!
!! TO DO improve MOD_PERSISTENT_CONTACTS module !!
    ---------------------------------------------------------------
  */

template<class P> 
double  
WallFeature::decision( P & p1, double overlap){

double normal_force = 0.;
double xy = material.getxy(p1.type,this->get_type()) ;
int g;

g = p1.stan;
if(overlap <= p1.xmax)
	{
	if(p1.xmax > xy)
		{
			double xp = compute_xp(p1,p1.xmax);
			if(overlap > xp)
			{
				//state 4
				normal_force = force_unloading_Hertz(p1,overlap,p1.xmax);
				//normal_force = force_loading_Hertz(p1,overlap);
				if(p1.stan == 4)
					{
						p1.control++;
					}
				else
					{
						p1.control = 0;
					}
				if(p1.stan != 4)
				p1.stan = 4;

			}
			else
			{
				//state 5	
				normal_force = 0.0;//force_unloading_Hertz(p1,overlap,p1.xmax);
				if (p1.stan != 5)
				p1.stan = 5;
			}
		}	
	else
		{
			//state no. 2
			normal_force = force_loading_Hertz(p1,overlap);
			if (p1.stan != 2)
			p1.stan = 2;
		}
	}
else
	{
		
		if(overlap > xy)
		{
			//state 3
			normal_force =plastic_force(p1,overlap) ;
			if (p1.stan != 3)
			p1.stan = 3;
		}
		else
		{
			//state 1
			normal_force = force_loading_Hertz(p1,overlap);
			if (p1.stan != 1)
			p1.stan = 1;
		}
	}
p1.overlap = overlap;
return normal_force;
};


//=================================================================

/** ---------------------------------------------------------------
    compute particle - particle force in plastic model
    ---------------------------------------------------------------
  */

template<class P> 
double WallFeature::plastic_force(P & p1, double overlap)
{
	double normal_force;
	normal_force =0.0 ;
	normal_force = material.getNForceYield(p1.type,this->get_type()) + M_PI*material.getpy(p1.type)*pow(material.getRR(p1.type,this->get_type()),2)*(overlap - material.getxy(p1.type,this->get_type()));
	return normal_force;
};


//=================================================================

/** ---------------------------------------------------------------
    compute particle - particle force in load Hertz model
    ---------------------------------------------------------------
  */

 template<class P> 
 double WallFeature::force_loading_Hertz(P & p1, double overlap)
{
	double normal_force; 
	normal_force = material.getForce(p1.type,this->get_type()) * overlap * sqrt(overlap);
	return normal_force;
};


//=================================================================

/** ---------------------------------------------------------------
    compute particle - particle force in unload Hertz model
    ---------------------------------------------------------------
  */

template<class P>
double WallFeature::force_unloading_Hertz(P & p1, double overlap,double xmax)
{
	double normal_force, xp, xe;
	xe = ( 3 * plastic_force(p1,xmax))/ (4 * material.getEE(p1.type,this->get_type()) * material.getRR(p1.type,this->get_type()));
	//xe = xe * sqrt(xe);
	xe = exp(log(xe)/3) * exp(log(xe)/3);
	xp = xmax - xe;
	//normal_force = (4 * material.getEE(p1.type,this->get_type()) * sqrt(material.getRR(p1.type,this->get_type())) * (overlap - xp) * sqrt(overlap - xp)) / 3 ;
	normal_force = material.getForce(p1.type,this->get_type()) * (overlap  - xp) * sqrt(overlap - xp);

	return normal_force;
};

//=================================================================

/** ---------------------------------------------------------------
    compute xp variable
    ---------------------------------------------------------------
  */

template<class P>
double WallFeature::compute_xp(P & p1,double  xmax)
{
	double xp, xe;
	xe = ( 3 * plastic_force(p1,xmax))/ (4 * material.getEE(p1.type,this->get_type()) * (material.getRR(p1.type,this->get_type())));
	//xe = xe * sqrt(xe);
	xe = exp(log(xe)/3) * exp(log(xe)/3);
	xp = xmax - xe;
	return xp ;
};



#if 0 
// unfinished work

class MDRotParticle;

class ContactVerification {
  // if the linked cell size is chosen to be equal to twice the maximum 
  // radius, then we might miss opening contacts in some cases (just imagine 
  // particles sitting within a distance from a linked cell boundary which 
  // corresponds to a distance smaller than that traversed within one
  // time step). Over time, the particle contact map might pollute with 
  // such bogus links. 
  // Fixing the problem either requires increasing the linked cell size 
  // [thus entailing a (cell-size increase)^3 run time penalty] or running 
  // a regular pairwise loop to identify the 'used' bonds and remove the
  // others. This is what we do here.

  // strategy: we identify all active pair contacts by running a 
  // loop over all particles and by inserting the overlapping particles into 
  // a list of contacts
  // we then run a loop over all particles and check if the contacts in 
  // their lists
  //   - correspond to one of the links found previously
  //   - or to a particle-wall contact
  // If neither is the case, we remove the dangling link 
  
  // use like this
  // ContactVerification   cv;
  // ...
  // cv.initialize( geometry );           // keeps track of walls as contact 
  // box.foreach_pair( cv.find() );       // find all contacts
  // box.foreach_particle( cv.check() );  // remove bogus contacts

private:
  
  struct CheckContacts {
    CheckContacts( SetType &a_set ): m_contacts(a_set); 
    void operator()( MDRotParticle &p );
    
    SetType & m_contacts;
  };
  
  struct FindContacts{
    FindContacts( SetType &a_set ): 
      m_contacts(a_set), m_material(a_mat); 
    void operator()( MDRotParticle &p1, MDRotParticle &p2  );
    
    SetType & m_contacts;
  };
  
  struct Cmp {
    int operator()( pair<unsigned long,unsigned long> a, 
                    pair<unsigned long,unsigned long> b ) {
      if      ( a.first < b.first ) 
	return -1;
      else if ( a.first > b.first ) 
	return 1;
      else if ( a.second < b.second ) 
	return -1;
      else if ( a.second > b.second )
	return 1;
      else return 0;
    }
  };

public:  
 
  FindContacts  find()  { return FindContacts(m_contacts); }
  CheckContacts check() { return CheckContacts(m_contacts); }
  
  void initialize(){ m_contacts.clear(); }
  
private:
  // active contacts
  typedef set< pair<unsigned long, unsigned long>, Cmp >  SetType; 
  SetType m_contacts;  
};
#endif

#endif  // include guard













