
#include "ModGeometry.H"
#include "PT/PTinttypes.h"

//template<class BASE, class TAGTYPE>
PTS_ExtTagFactory<WallFeature,std::string> Geometry::our_factory;


void 
Geometry::set_wallfeature_ids(Control &ctl, PTS_CommEnv &ce){

  for ( vector<WallFeature *>::iterator it = allWalls.begin();
	it != allWalls.end();
	++it ) {
    
    uint32_t id;
    if ( ce.node_id() == 0 )
      id = ctl.unique();
    ce.broadcast(id);   // everybody has it now
    
    (*it)->set_id(id);
    PT_DEBUGINFO3( "using wallfeature id " << id );  
    
  } 

}


void 
Geometry::select_features(double dist, const double_vec proc_ll, const double_vec proc_ur){

  // for all features in the list, we decide whether a particle in 
  // the volume specified by ll and ur can possibly have an interaction 
  // with it. If yes, the feature is added to a 'selected' list 
  // and only that list is scanned when operator() is called to 
  // compute particle-wall interactions
    
  localWalls.clear();
  for ( vector<WallFeature *>::iterator it = allWalls.begin();
	it != allWalls.end();
	++it ) {
    double_vec  l_ll, l_ur;
    (*it)->region(dist, l_ll, l_ur);
    // I think one must check whether there are overlaps of all the 
    // intervals ur-ll of the regions of interest
    bool overlap = true;
    for ( int dim=0; dim < DIM; ++dim )
      if ( (proc_ll[dim] <= l_ur[dim]) && (l_ll[dim] <= proc_ur[dim] )) // overlap 
	continue;
      else {
	overlap = false;
	break;
      }
    if ( overlap )
      localWalls.push_back( *it ); // add pointer to the list
    
  } // all features
  PT_DEBUGINFO( "added " << localWalls.size() << " of " 
		<< allWalls.size() << " features to local list" ); 
}


int 
Geometry::size()
{ 
  return allWalls.size(); 
}

Geometry::Geometry( istream &in )
{
  init(in);
}

Geometry::Geometry( PTC_Para &pf )
{
  string strProject = 
    pf.getd("project_name","no_project");  // project name
  string strGeometry=strProject+".geo";
  pf.get("geometry_name",strGeometry);    //geometry file
  PT_DEBUGINFO("geometry file=" << strGeometry ); 
  
  std::ifstream  in(strGeometry.c_str());
  init(in);
  PT_INFO( "geometry with " << allWalls.size() << " objects" );
}
  
void 
Geometry::update( double t )
{
  // must iterate over all walls because the statistics might require 
  // equal 'configuration' counts on all processors. This count is 
  // typically incremented in the update routines
  for ( vector<WallFeature *>::iterator it = allWalls.begin();
	it != allWalls.end();
	++it ) {
    (*it)->update(t);  // either updates statistics or modifies configuration
  }
}

void 
Geometry::init( istream &in)
{   
  // must make sure that material is initialized, otherwise doom looms
  if ( WallFeature::material.size() <= 0 ) 
    PT_SEVERE( "static 'material' variable of WallFeature is in "
	       "inconsistent "
	       "state, needs to be initialized by assignment prior to "
	       "Geometry ctor call" );
  
  static int invocations;
  if ( invocations == 0 ){
    our_factory.reg<Rectangle>("wall");
    our_factory.reg<ConvexFeature>("convex1");
    our_factory.reg<Cylinder>("cylinder");
    our_factory.reg<WallSphere>("sphere");
    our_factory.reg<MovingRectangle>("moving_wall");
    invocations++;
  }
  // read in the stuff from the parameter file
  
  std::string line;
  do { 
    getline(in,line);
    std::string::size_type pos = line.find_first_not_of(" \t\n");
    if ( pos == std::string::npos ) 
      continue; 
    else if ( line[pos] == '!' ) // comment
      continue;
    else {
      
      PT_DEBUGINFO( line );
      istringstream strin(line);
      std::string key;
      strin >> key; 
      
      if ( key == "box" ) 
	strin >> ll >> ur;
      else {
	PT_ASSERT(    key == "wall" 
		   || key == "convex1" 
		   || key == "cylinder" 
		   || key == "sphere"	
		   || key == "moving_wall" );
	allWalls.push_back( our_factory.make(key,strin) );
      }
    } 
    
  } while ( in.good() );
}

std::ostream &
Geometry::save(std::ostream &out) const 
{
  for( std::vector<WallFeature *>::const_iterator it=allWalls.begin();
       it != allWalls.end(); ++it ){
    // call wall's binary save routine
    (*it)->save(out);  // essentially saves the statistics 
  }
  return out;
}    

std::istream &
Geometry::restore(std::istream &in) 
{
  for( std::vector<WallFeature *>::iterator it=allWalls.begin();
       it != allWalls.end(); ++it ){
    // call wall's binary save routine
    (*it)->restore(in);  // essentially saves the statistics 
  }
  return in;
}    

void Geometry::ModIO_ascii_in(std::istream &in)
{
  PT_INFO( "ascii input of geometry feature data not supported, skipping" );
  // search for end marker as often as necessary, i.e. for as many as we have 
  // geometry elements
  int count = 0;
  std::string line;
  do {
    std::getline(in,line);
    if ( line == "### END GEOMETRY FEATURE DATA ###" ) count++;
  }
  while ( count < static_cast<int>( allWalls.size() ) );

  PT_DEBUGINFO( "status before return: " << (bool) in );
}

void 
Geometry::ModIO_ascii_out(PTS_CommEnv &ce, int root_node, std::ostream &out)
{
  for ( std::vector<WallFeature *>::iterator it = allWalls.begin();
	it != allWalls.end(); ++it ) {

    if ( root_node == ce.node_id() ) {
      (*it)->reduce().print(out);
      out << "### END GEOMETRY FEATURE DATA ###\n";
    }
    else
      (*it)->reduce();

    (*it)->clear();
  }
}

void 
Geometry::ModIO_bin_in(std::istream &in)
{
  for ( std::vector<WallFeature *>::iterator it = allWalls.begin();
	it != allWalls.end(); ++it ){
    (*it)->restore(in);
  }
}

void 
Geometry::ModIO_bin_out(PTS_CommEnv &ce, int root_node, std::ostream &out)
{
  for ( std::vector<WallFeature *>::iterator it = allWalls.begin();
	it != allWalls.end(); ++it ){

    if ( ce.node_id() == root_node ) {
      (*it)->reduce().save(out);
    }
    else 
      (*it)->reduce();

    (*it)->clear();
  }
}


#ifdef PTM_VISUALIZE
  // draws geometry
void 
Geometry::draw(XTDisplay &window)
{
  std::vector<WallFeature*>::iterator it;
  
  // loop over all walls in list
  for( it=allWalls.begin();it != allWalls.end(); ++it ){
    // call wall's drawing routine
    (*it)->draw(window);
  }
}
#endif
 
















