// $Id: ModIO.cc,v 1.6 2001/09/02 20:54:55 sts Exp $

#include "ModIO.H"
#include "PT/PTS_ParStream.h"
#include "PT/PTBinaryIO.h"
  
//  ================  WRITE SAVE FILE ================

//  -- what will be written as per-particle data by this routine is 
// defined by member functions of the particle class in ModParticle.H
// and ModGeometry.H (which might wnat to save per-wall data)
// check and change those if you need more or other data in the 'save' files
bool
WriteSaveFile(ostream &out, double t, ParticleBox &box, Geometry &geo, 
	      PTS_CommEnv &ce, PTC_Para &pf ){

  // write configuration files. If binary is true, in binary form, else in 
  // ASCII (for debugging or small systems). If you need to extend the 
  // scheme, make sure to save the exact same data on CPU 0 and all others

  bool binary = pf.getd("use_binary_save_file",true); 
  if ( binary ) set_binary_format(out, pf, "binary_save_format");

  // write CPU 0s stuff
  if ( ce.node_id() == 0 ) {

    if ( binary ){ 
      out < t < ce.n_nodes();
      out < 0 < box.size();
    } 
    else { 
      out << "#_t=  " << t << " ncpus= " << ce.n_nodes() << '\n';
      out << "#_cpu " << 0 << " n_p= " << box.size() << '\n';  
    }
    for ( ParticleBox::iterator it= box.begin();
	  it != box.end();
	  ++it ){

      if ( binary ) {
	it->ModIO_bin_out(out);
      }
      else {
	it->ModIO_ascii_out(out);
      }
    }

  }

  // on  node 0 we run a loop over all CPU's, 
  // broadcast the Id from where we would like to receive particles
  // then receive the particles until eof /  the alternative is 
  // to use barrier calls to let other CPU's wait for the p2p 
  // communication to finish  
  for ( int id=1; id < ce.n_nodes(); ++id ){

    int rid = id;
    ce.broadcast(rid); // root is default 0 
    PT_ASSERT( rid > 0 );
    PT_ASSERT( rid < ce.n_nodes() );

    // on all other nodes, we evaluate the broadcast value
    // if it matches our id, its our turn to send
    if ( ce.node_id() == rid ){   // our turn
      
      // name hiding intended (out is the output file in enclosing scope)
      PTS_ParStream   channel(ce,0);  // talk to 0

      if ( binary ) 
	channel < box.size();
      else
	channel << box.size() << " ";

      for ( ParticleBox::iterator it= box.begin();
	    it != box.end();
	    ++it ){

	if ( binary ) 
	  it->ModIO_bin_out(channel);
	else 
	  it->ModIO_ascii_out(channel);
      }

    } // destroys stream = terminates communication
    else if ( ce.node_id() == 0 ){ // on receiving 'master' node 
      
      PTS_ParStream   channel(ce,id);  // receive stuff from id

      int size;
      if ( binary ){
	channel   > size;
	out < rid < size;
      }
      else { 
	channel >> size;
	out << "#_cpu  " << rid << " n= " << size << '\n';  
      }

      while ( channel ) {

	if ( binary ) {
	  ParticleBox::value_type  p;
	  p.ModIO_bin_in(channel);
	  if ( !channel.fail() ){
	    p.ModIO_bin_out(out);
	  }
	}
	else {
	  ParticleBox::value_type p;
	  p.ModIO_ascii_in(channel);
	  if ( !channel.fail() ){
	    p.ModIO_ascii_out(out);
	  }
	}
      }
    } // closing channel

    // otherwise it's not our turn and we wait in the next broadcast.

  } // for ( id )

  // after the particle info we write the geometry/wall data
  // need to call this on all CPUs because communication between statistics 
  //  data members is required and would otherwise deadlock  
  if ( binary )
    geo.ModIO_bin_out(ce,0,out);
  else 
    geo.ModIO_ascii_out(ce,0,out); 

  out.flush();   // try to get everything into the save file

  return out;
}    


bool 
ReadSaveFile(istream &in, double &t, ParticleBox &box, Geometry &geo, 
	     PTC_Para &pf ){

  // reading takes place on one CPU and it consumes the amount of 
  // data that has been written by one call to WriteSaveFile before
  // the setting of binary_save_format and use_binary_save_file 
  // are respected and must equal the values used in the simulation

  // box si supposed to be empty and is filled after the call
  // t will be set from the value in the file
  // pf will be queried for binary_save_format and use_binary_save_file

  // format  
  // time   tot_n_particles
  // CPU_id n_particles
  // particle data

  int n_particles_read_cpu = 0;
  int n_particles_read_total = 0;
  int n_particles_cpu      = 0;
  int n_cpus;
  int cpu_id;
  int cpu_count =0;
  string dummy;

  bool binary = pf.getd("use_binary_save_file",true); 
  if ( binary ) set_binary_format(in, pf, "binary_save_format");

  PT_INFO( " input status (A0) " << (bool) in );
  if ( binary )
    in > t > n_cpus;
  else {
    in >> dummy >> t;
    cout << dummy << " " << t << "\n";
    in >> dummy >> n_cpus;
    cout << dummy << " " << n_cpus << "\n";
  }
  PT_INFO( " input status (A1) " << (bool) in );
  if ( !in ) {
    return in; // error or eof
  }  
  PT_INFO( "t = " << t << ", N_CPUs = " << n_cpus );

  while ( cpu_count < n_cpus ) {

    cpu_count++;
    n_particles_read_cpu = 0;
    if ( binary )
      in > cpu_id > n_particles_cpu;
    else 
      in >> dummy >> cpu_id >> dummy >> n_particles_cpu;
  
    while ( n_particles_read_cpu < n_particles_cpu ){

      ParticleBox::value_type p;

      if ( binary ) {
	p.ModIO_bin_in(in);
      }
      else {
	p.ModIO_ascii_in(in);
      }
      PT_DEBUGINFO( p );
      box.add(p);
      n_particles_read_cpu++;
      n_particles_read_total++;
    }
    
  }
  PT_INFO( "in total N = " << n_particles_read_total << " particles" );

  box.add_end();

  // geometry/wall (force) data
  if ( binary )
    geo.ModIO_bin_in(in);
  else
    geo.ModIO_ascii_in(in);

  PT_INFO( " input status (B) " << (bool) in );

  return in; // state 
}




// ----------------------------------------------------------------------

int 
set_binary_format(std::istream &istr, PTC_Para &pf, string keyword){

  // by default, use CRAY format for binary I/O
  std::string format = pf.getd(keyword,"cray-t3e"); 
  
#if PT_CONVERTING_BINARYIO
  istr > PT_BinaryIO::stream_format(format);
#endif
  return 0;
}

int 
set_binary_format(std::ostream &ostr, PTC_Para &pf, string keyword){

  // by default, use CRAY format for binary I/O
  std::string format = pf.getd(keyword,"cray-t3e"); 
  
#if PT_CONVERTING_BINARYIO
  ostr < PT_BinaryIO::stream_format(format);
#endif
  return 0;
}


int    
WriteRestartFile(ofstream & out, ParticleBox & box, Geometry &geo, 
		 Control & control, double t, int t_step, 
		 PTC_Para &pf )
{
  set_binary_format(out, pf, "binary_restart_format");

  // if not at beginning position at the beginning
  out.seekp(0);
  out < box < geo < control < t < t_step;
  // make sure it's on disk:
  out.flush();
  return out.good();
}


int    
ReadRestartFile(istream & restart_in, ParticleBox & box, Geometry &geo, 
		Control & control, double & t, int & t_step, 
		PTC_Para &pf )
{
  set_binary_format(restart_in, pf, "binary_restart_format");

  if ( restart_in ) {
    restart_in > box > geo > control > t > t_step; 
    if ( restart_in.good() ) {
      PT_INFO( "read restart file" );
      PT_INFO( "simulation time (/-step): " << t << " / " << t_step );
      // quick test for consistency - all t_step must be equal
#if PT_PARALLEL == 1
      int t_step_min=0, t_step_max=42;
      PT_VERIFY(MPI_SUCCESS==(MPI_Allreduce(&t_step,&t_step_min,1,MPI_INT,MPI_MIN,MPI_COMM_WORLD)));	
      PT_VERIFY(MPI_SUCCESS==(MPI_Allreduce(&t_step,&t_step_max,1,MPI_INT,MPI_MAX,MPI_COMM_WORLD)));	
      if( t_step_min != t_step_max ) {
	PT_SEVERE( "inconsistent restart files, save time step differs" );
	return 1;
      }
#endif		
      box.update();
      return 0;
    }
    else {
      // inconsistent state
      PT_SEVERE( "failed to initialize from restart file" );
      return 1;
    }
  } 
  else { 
    // missing restart file is probably ok
    PT_INFO("couldn't open restart file" );
    return 1;
  }
}


// ----------------------------------------------------------------------

int
ProgressInfo(ostream &out, ParticleBox & box, 
	     double t, int t_step, PTS_Resource &res )
{
  static double last_now;
  static int    last_step;

  out << "t_sim=" << t << ", N_p=" << box.global_size();
  if ( last_step != 0 && last_step != t_step && box.size() > 0 )
    out << ", us/part=" 
	<< (1e6 * (res.now() - last_now) / box.size() / (t_step-last_step));
  out << std::endl;

  last_now  = res.now();
  last_step = t_step;

  return 0;
}


// ======================== ball file =================================


int 
WriteBallFile(ostream &out, ParticleBox &box)
{
  // create from the current particle content of 'box' a file that can 
  // serve as 'ball file' in a ballfile control event for loading an
  // initial configuration/a number of particles at defined positions

  out << "! ball configuration N = " << box.size() << " particles\n"
         "!\n"; 
  
  for ( ParticleBox::iterator it = box.begin();
	it != box.end();
	++it ) {
    
    out << "ball " << (*it).x << " " << (*it).v 
	<< " " << static_cast<int>((*it).type) 
	<< "\n";
    
  }
  out << std::endl;
  return box.size();
}







