#ifndef MOD_MATERIAL_H
#define MOD_MATERIAL_H

#include "PT/PTmath.h"
#include "IOFuncs.H"
#include "PT/PTC_Para.h"
#include "PT/PTS_IArray.h"

#include <vector>
#include "ModParticle.H"  // need double_vec 

// there creep up instantiation problems on the Cray if
// normal_friction_switch is vector<vector<bool>> 
// we thus use <int> on the Cray:
#if PT_CRAY_T3E
 typedef int MOD_VECBOOL;
#else 
 typedef bool MOD_VECBOOL;
#endif

struct Mat {
  double E;
  double nue;
  double B;
  double sigma;
  double density;
  double r;
  double color;
  double py;
  double mass;           // computed 
  double mom_of_inertia; // computed
};


std::istream & operator>>(std::istream &in, Mat &mat);


class Material {
private:

  Material(){}
  friend class WallFeature;

  std::vector<Mat>     mat;
  typedef PTS_IArray<double,2>  array2d;
  array2d kDamp, kForce, kFric,RR,EE,xy;
  array2d tangDamp;             // stores precomputed value for the 
                                // "critical" tangential damping 
  array2d NForceYield; 		// Normal force at yielding 
  std::vector< std::vector<double> > mue;
  std::vector< std::vector<MOD_VECBOOL> >   normal_friction_switch;
  double  timestep_gpk5;
  double  cutoff;
  double_vec gravity;
  // bools to trace if a value was found in the materials file
  bool   gravity_found;
  bool   normal_friction_switch_found;
  bool   mue_found, timestep_gpk5_found, cutoff_found; 

  void   precalculate();
  double min_cutoff() const; // min necessary cutoff from particle radii
  
public:

  Material(PTC_Para &pf);
  
  double get_mue(int i, int j) const { 
    return mue[i][j]; 
  }

  /** return an appropriate time step value for the integrator */
  double get_timestep_gpk5() const { return timestep_gpk5; }

  // interaction range cutoff for cell size det. in container
  double get_cutoff() const { return cutoff; }

  // Hertz prefactor
  double getForce(int i, int j) const {
    return kForce[i][j];
  }
 
  // Normal Force Yield prefactor
  double getNForceYield(int i, int j) const {
  return NForceYield[i][j];
}

  // wytrzymalosc na sciskanie prefactor
  double getpy(int i) const {
  return mat[i].py;
}

  // overlap start plastic deformation  prefactor
  double getxy(int i, int j) const {
  return xy[i][j];  
}



  //Radius two spheres prefactor 
  double getRR(int i, int j) const {
    return RR[i][j];
  }
 
  // E. two spheres prefactor
  double getEE(int i, int j) const {
    return EE[i][j];
  }
 
  // non linear damping prefactor
  double getDamp(int i, int j) const {
    return kDamp[i][j];
  }

  // damping of the tangential force spring
  double getTDamp(int i, int j) const {
    return tangDamp[i][j];
  }

  // static friction spring prefactor
  double getFric(int i, int j) const {
    return kFric[i][j];
  }

  double getRadius(int i) const {
    return mat[i].r;
  }

  double getMass(int i) const {
    return mat[i].mass;
  }

  double_vec getWeight(int i) const {
    return mat[i].mass * gravity;
  }

  double getMomentOfInertia(int i) const {
    return mat[i].mom_of_inertia;
  }

  double getColor(int i) const {
    return mat[i].color;
  }

  // number of material parameter sets
  int    size(void) const {
    return mat.size();
  }

};

#endif










