#include "ModMaterial.H"
#include "PT/PTDebug.h"
#include <math.h>

std::istream & 
operator>>(std::istream &in, Mat &mat){
  in >> mat.E >> mat.nue >> mat.B >> mat.sigma 
     >> mat.r >> mat.density; 
  PT_VERIFY( in );
  in >> mat.color; // might encounter eof
  in >> mat.py;
  if ( mat.nue > 0.5 ) // for thermodynamic reasons
    PT_WARNING( "poisson ratio > 0.5 impossible" );
  if ( mat.E < 0. )
    PT_WARNING( "Young modulus negative" );
  if ( mat.r <= 0. )
    PT_WARNING( "particle radius negative or zero" );
  if ( mat.density <= 0. )
    PT_WARNING( "density less or equal zero" );
    
  mat.mass           = 4./3. * M_PI * mat.r * mat.r * mat.r * mat.density; 
  mat.mom_of_inertia = (2./5.) * mat.mass * mat.r * mat.r;
//  std::cout<<"density "<<mat.density<<std::endl;
  return in;
}


double Material::min_cutoff() const {
  double min_cutoff = 0.;
  // the minimum necessary interaction range
  // which in our case is twice the largest radius
  for( unsigned i=0; i < mat.size(); ++i ){
    if ( mat[i].r * 2. > min_cutoff ) 
      min_cutoff = mat[i].r * 2.;
  }
  return min_cutoff;
}


void Material::precalculate() {

  // give arrays the necessary size
  int size[] = { mat.size(), mat.size() };
  kFric    = array2d(size);
  kDamp    = array2d(size);
  kForce   = array2d(size);
  tangDamp = array2d(size);
  RR 	   = array2d(size);
  EE       = array2d(size);
  xy 	   = array2d(size);
  NForceYield = array2d(size);
  for ( unsigned j=0; j<mat.size(); j++) {
    for ( unsigned i=0; i<mat.size(); i++) {
      double rr = sqrt(mat[i].r*mat[j].r/(mat[i].r+mat[j].r));
      RR [i][j] = rr;
      kDamp [i][j] = (2. * mat[i].B * mat[j].B) * rr / 
       (mat[i].B*(1-mat[j].nue*mat[j].nue)+mat[j].B*(1-mat[i].nue*mat[i].nue));
      
      // we'll mask the damping coefficients by normal_friction_switch
      kDamp [i][j] = normal_friction_switch[i][j] ? kDamp [i][j] : 0.;
      if ( kDamp[i][j] == 0. )
	PT_WARNING( "no dissipation between materials " 
		    << i << " and " << j);  
      if ( normal_friction_switch[i][j] != normal_friction_switch[j][i] )
	PT_WARNING( "inconsistent normal friction switch " << i << "/" << j );
      if ( mue[i][j] != mue[j][i] ) 
	PT_WARNING( "mue is not a symmetric matrix " << i << "/" << j ); 
      if ( mue[i][j] < 0 )
	PT_WARNING( "negative friction coefficient, material " << i );
    /*  EE[i][j] = ( mat[i].E * mat[j].E)  /
         ( mat[i].E*(1-mat[j].nue*mat[j].nue) 
                +mat[j].E*(1-mat[i].nue*mat[i].nue)) ;
	
      kForce[i][j] = (4. * mat[i].E * mat[j].E) * rr /
        ( 3. * ( mat[i].E*(1-mat[j].nue*mat[j].nue) 
                +mat[j].E*(1-mat[i].nue*mat[i].nue)) );
      */
	EE[i][j] = ( mat[i].E * mat[j].E)  /
         ( (1-pow(mat[i].nue,2))*mat[j].E + (1-pow(mat[j].nue,2))*mat[i].E) ;
double sprawdz =EE[i][j];
	
      kForce[i][j] = (4. * mat[i].E * mat[j].E) * rr /
        ( 3. * ( mat[i].E*(1-pow(mat[j].nue,2))+ mat[j].E*(1-pow(mat[i].nue,2))));


      double G1 = mat[i].E / (2. * (1+mat[i].nue) );
      double G2 = mat[j].E / (2. * (1+mat[j].nue) );
      kFric[i][j] = (16.*G1*G2) * rr / 
	            (3.*(G1*(2-mat[j].nue)+G2*(2-mat[i].nue)));

      // estimate critical tangential damping (factor h^1/4 added later) 
      // multiply gamma with the reduced mass here to save that step in the
      // force routine. That yields a square root( mass ) in the numerator 
      tangDamp[i][j] = sqrt( 8./7. * kForce[i][j] * 
		  (mat[i].mass * mat[j].mass) / (mat[i].mass + mat[j].mass) );
	
	xy[i][j] = pow((M_PI * mat[i].py)/( 2 * getEE(i,j)),2) * pow(getRR(i,j),2);
        NForceYield[i][j] = 2 * M_PI * pow(getRR(i,j),2) * xy[i][j] * (mat[i].py/3);
	/* else {  
	 if(i ==0)
		{
		 xy[i][j] = pow((M_PI * mat[j].py)/( 2 *mat[j].E),2) *  pow(getRR(i,j),2);
//getRadius(j);
       		 NForceYield[i][j] = 2 * M_PI *  * xy[i][j] * (mat[j].py/3);
		}
     	 else
		{
			if(j ==0)
			{
			 xy[i][j] = pow((M_PI * mat[i].py)/( 2 * mat[i].E),2) *  pow(getRR(i,j),2);
//getRadius(i);
               		 NForceYield[i][j] = 2 * M_PI *  pow(getRR(i,j),2)* xy[i][j] * (mat[i].py/3);
			}
			else
			{
  			    xy[i][j] = pow((M_PI * mat[i].py)/( 2 * getEE(i,j)),2) * pow(getRR(i,j),2);
                   	    NForceYield[i][j] = 2 * M_PI * pow(getRR(i,j),2) * xy[i][j] * (mat[i].py/3);
			}
		}
	     }*/	
 	   }
  	}
  	
}
  
Material::Material(PTC_Para &pf)    
: timestep_gpk5(0.00001), cutoff(2.), 
  gravity(0.,0.,-9.81), gravity_found(false),
  normal_friction_switch_found(false),
  mue_found(false), timestep_gpk5_found(false), cutoff_found(false)
{
  string strMat = pf.getd("material_file",""); //name of material file

  ifstream tFile(strMat.c_str());
  if ( !tFile ){
    PT_SEVERE( "couldnt open materials file " << strMat );
    exit(-1);
  }
  
  while (tFile){
    
    string strLine, strWord;
    getline(tFile,strLine);
    
    // Is there a comment in this line?
    string::size_type tPos = strLine.find_first_of('!');
    if (tPos!=string::npos){
      // Remove comment
      strLine=strLine.erase(tPos);
    }
    
    // Get command string
    strWord=ParseString(strLine);
    strWord=LowerCase(strWord);
    
    if (strWord=="mat"){
      Mat        temp_mat;
      std::stringstream str_in(strLine);
      str_in >> temp_mat;
      mat.push_back(temp_mat);
    }
    else if (strWord=="mue"){
      // a single item in strLine is interpreted as a global value for 
      //  mue that fits all material pairs 
      // otherwise, we expect a symmetric matrix of values, delimited by []
      //  it must be given on one line!
      string::size_type pos = strLine.find_first_of("[");
      if ( pos == PT_STR_NPOS ){
	double tmp_mue = getDouble(strLine);
	for ( unsigned i=0; i < mat.size(); ++i ){
	  mue.push_back( std::vector<double>() );
	  for ( unsigned j=0; j < mat.size(); ++j )
	    mue[i].push_back( tmp_mue );
	}
      }
      else {
	stringstream in_str(strLine);
	in_str >> mue;
	if ( mue.size() != mat.size() ){
	  PT_SEVERE( "matrix of mu's does not have as many entries as "
		     "material types exist\n" );
	}
      }
      mue_found = true;
    }
    else if (strWord == "normal_friction_switch" ) {

      normal_friction_switch_found = true;

      // bool/int values (should be bool, but introduced int to fight 
      //  a KCC 3.3d instantiation problem).
      // must come after material definitions
      // must be symmetric
      // must have N_mat x N_mat entries 

      stringstream in_str(strLine);
      in_str >> normal_friction_switch;
      if ( normal_friction_switch.size() != mat.size() ){
	PT_SEVERE( "matrix of friction switches does not have as many "
		   "entries as material types exist\n" );
      }

    }
    else if (strWord=="timestep_gpk5"){
      timestep_gpk5 = getDouble(strLine);
      timestep_gpk5_found = true;
    }
    else if (strWord=="timestep"){
      
      timestep_gpk5 = getDouble(strLine);
      timestep_gpk5_found = true;
    }
    else if (strWord=="cutoff"){
      
      cutoff = getDouble(strLine);
      cutoff_found = true;
    }
    else if ( strWord=="gravity" ){

      stringstream in_str(strLine);
      in_str >> gravity;
      gravity_found = true;

    }
    else if (strWord==""){
      // do nothing
    }
    else {
      PT_WARNING( "materials file, unknown keyword " << strWord );
    }
  } // while ( file )

  if ( !normal_friction_switch_found )
    normal_friction_switch = std::vector<std::vector<MOD_VECBOOL> >
			 (mat.size(), std::vector<MOD_VECBOOL>(mat.size(), 1));

  if ( mat.size() > 0 )
    precalculate();

  if ( gravity_found == false ) 
    PT_WARNING( "no gravity specified in materials file, using "
		"default value " << gravity ); 

  double tmp_cutoff = this->min_cutoff();
  if ( cutoff_found ) {     // cutoff from materials file
    if ( tmp_cutoff > cutoff ){
      PT_WARNING( "explicit cutoff " << cutoff << " smaller than "
		  "max. interaction range; "
                  "ignore warning if due to large radius in wall "
		  "material specification" );
    }
  }
  else 
    cutoff = tmp_cutoff;

  PT_INFO( "cutoff: " << cutoff );
  PT_INFO( "time step: " << timestep_gpk5 );
  PT_INFO( "gravity: " << gravity );

}















