#ifndef MODPARTICLE_H
#define MODPARTICLE_H

#include "PT/PTBinaryIO.h"
#include "PT/PTS_CMapVector.h"

#include "PT/PTZ_Vector.h"
static   const int DIM = 3; 
typedef  PTZ_Vector<double,DIM> double_vec;
typedef  PTZ_Vector<float,DIM>  float_vec;
typedef  PTZ_Vector<int,DIM>    int_vec;

#include "ModMaterial.H"

/** class with required data members 
    must meet container requirements:
    - x must be public data member
    - flat copy must be enough
 **/

// ---------------------------------------------
// particle class (for GEAR 5 pred.-corr.)
// ---------------------------------------------
#ifndef MOD_JANSEN_PRESSURE         // requires additional time and space
#define  MOD_JANSEN_PRESSURE 0
#endif

#ifndef MOD_FORCES_IN_SAVE_FILE
# define MOD_FORCES_IN_SAVE_FILE 0
#endif

#if MOD_FORCES_IN_SAVE_FILE && MOD_JANSEN_PRESSURE 
#error can only specify one of MOD_FORCES_IN.. and MOD_JANSEN_PRESSURE
#endif 

#ifndef MOD_PERSISTENT_CONTACTS     
# define MOD_PERSISTENT_CONTACTS 1
#endif

#ifndef MOD_ROTATION
# define MOD_ROTATION 1
#endif

#ifndef MOD_CMAP_SORTED
# define MOD_CMAP_SORTED 0
#endif
#if MOD_CMAP_SORTED == 0 
# include "PT/PTS_CMapVector.h"
#else 
# include "PT/PTS_CMapSorted.h"
#endif


#if MOD_PERSISTENT_CONTACTS
const int MAX_CONTACTS = 20;
class MDParticle;

struct MDContact {
  MDContact(const double_vec& ds, const double_vec& dot_ds/*, double_vec& force*/) : 
    delta_s(ds), delta_s_dot(dot_ds), active(true) , overlap_max(0.)/*, Contact_force(force)*/{}
  MDContact() : delta_s(0.), delta_s_dot(0.), active(true), overlap_max(0.)/*, Contact_force(0.)*/ {}

  // contact data:
  double_vec delta_s;     // Lagrange multiplyer to find static contact force  
  double_vec delta_s_dot; // velocity of change
  bool active;
  double overlap_max;
  double_vec Contact_force;
  // set to true if encountered in particle contact check
  // if false after the time step, the contact can be removed entirely

  // write contact data, restricted to 'meaningful' data
  ostream &save( ostream &out ) const;
  // read contact data, ... 
  istream &restore( istream &in );
  void ModIO_ascii_out( ostream &o ) const {
    //o << Contact_force
    o<< " " << overlap_max << '\n'; 
  }

};
#endif

class MDParticle {
public:
  double_vec x;   // position
  double_vec v;   // velocity
  double_vec a;   // acceleration
  double_vec x3;  //
  double_vec x4;  //
  double_vec f;   // force

  unsigned char type;
  unsigned long id;   // globally unique particle id
  int stan,control;
  double xmax; //tmp
  double overlap; //tmp

#if MOD_PERSISTENT_CONTACTS
  // maps particle ids to contacts
# if MOD_CMAP_SORTED == 1
  PTS_CMapSorted<unsigned long,MDContact,MAX_CONTACTS> contact;
  typedef PTS_CMapSorted<unsigned long,MDContact,MAX_CONTACTS>::iterator CMapIterator; 
# else
  PTS_CMapVector<unsigned long,MDContact,MAX_CONTACTS> contact;
  typedef PTS_CMapVector<unsigned long,MDContact,MAX_CONTACTS>::iterator CMapIterator; 
# endif
#endif

  //  double     m;   // mass
  //  double     r;   // radius
  //  double     color;     

  MDParticle (): 
    x(0.), v(0.), a(0.), x3(0.),
    x4(0.), f(0.), type(0), id(0), stan(0), xmax(0.0),overlap(0.0)
  {}
  
  // write particle data to stream
  ostream &save( ostream &out ) const;

  // read particle data from stream
  istream &restore( istream &in );
};

struct MDRotParticle: public MDParticle {

  double_vec omega;  // angular velocity
  double_vec alpha;  // ang acc.
  double_vec omega2; // omega''
  double_vec omega3; 
  double_vec omega4; 
  double_vec torque; 
#if MOD_JANSEN_PRESSURE == 1
  double   pressure;  // sum of absolute values of normal forces
#endif

  MDRotParticle (): 
    omega(0.), alpha(0.), omega2(0.), omega3(0.), omega4(0.), 
    torque(0.)
#if MOD_JANSEN_PRESSURE == 1
    , pressure(0.)  
#endif
  { }

  // write particle data to stream
  ostream &save( ostream &out ) const;

  // read particle data from stream
  istream &restore( istream &in );

  // routines to output restricte particle info in the context of the 
  // WriteSaveFile routine in ModIO.H 
  void ModIO_bin_out( ostream &o) const  {
    // we write the particle info as float to save space
    o < float_vec(x) < float_vec(v) < float_vec(omega) 
#if MOD_FORCES_IN_SAVE_FILE
      < float_vec(f) < float_vec(torque) 
#endif
#if MOD_JANSEN_PRESSURE
      < pressure
#endif
      < type < id;
  }

  // and the reverse input function (note that the binary data must be read 
  // exactly as written, yoiu cannot use the particle member variables directly
  // because their type doesn't match the one used for writing
  void ModIO_bin_in( istream &i ) {
#if MOD_FORCES_IN_SAVE_FILE
    float_vec fx, fv, fomega, ff, ftorque; 
    i > fx > fv > fomega > ff > ftorque > type > id;
    x = fx; 
    v = fv;
    omega = fomega;
    f = ff;
    torque = ftorque;
#else
    float_vec fx, fv, fomega;
    i > fx > fv > fomega 
#if MOD_JANSEN_PRESSURE
      > pressure
#endif
      > type > id;
    x = fx; 
    v = fv;
    omega = fomega;
#endif
  }

  // particle data to be saved in ascii mode 
  void ModIO_ascii_out( ostream &o ) const {
    o << x << v <<omega  
#if MOD_FORCES_IN_SAVE_FILE
    << f << torque 
#endif
#if MOD_JANSEN_PRESSURE
      << pressure
#endif
      << " " << (int) type << " " << id << '\n';;
      
      
  }

  // reverse 
      
  void ModIO_ascii_in( istream &i ) {
    int in_type;
    i >> x >> v >> omega 
#if MOD_FORCES_IN_SAVE_FILE
      >> f >> torque 
#endif
#if MOD_JANSEN_PRESSURE
      >> pressure
#endif
      >> in_type >> id;
    PT_ASSERT(in_type >=0 && in_type < 255);  // type is originally unsigned char
    type = in_type;
  }

};

// needs binary i/o for restore/save
// binary output
inline ostream & MDParticle::save(ostream & out) const {
  out < x < v < a < x3 < x4 < type < id;
#if MOD_PERSISTENT_CONTACTS
  out < contact;
#endif
  return out;
}

// binary input
inline istream & MDParticle::restore(istream & in) {
  in > x > v > a > x3 > x4 > type > id;
#if MOD_PERSISTENT_CONTACTS
  in > contact;
#endif
  return in;
}

// needs binary i/o for restore/save
inline ostream & MDRotParticle::save(ostream & out) const {
  MDParticle::save(out);
  return out < omega < alpha < omega2 < omega3 < omega4 < torque;
}

// needs binary i/o for restore/save
inline istream & MDRotParticle::restore(istream & in) {
  MDParticle::restore(in);
  return in > omega > alpha > omega2 > omega3 > omega4 > torque;
}


#if MOD_PERSISTENT_CONTACTS
// needs binary i/o for restore/save
// binary output
inline ostream & 
MDContact::save(ostream & out) const {
  return out < delta_s < delta_s_dot;
}

// binary input
inline istream & 
MDContact::restore(istream & in) {
  return in > delta_s > delta_s_dot;
}
#endif

// ascii output
inline ostream & operator<<(ostream & out, const MDParticle & p){
  out << p.x << p.v; // symmetric with input << p.a << p.x3  << p.x4  << p.f;
  out << (int) p.type;
  return out;
}

// ascii input
inline istream & operator>>(istream & in, MDParticle & p )
{
  int type;
  in >> p.x >> p.v >> type; 
  p.type = type;
  PT_SOMECHECKS( 
  if( !in ) 
    PT_WARNING( "particle specification incorrect: [pos] [vel] type" );
  )
  return in;
}

// ---------- APPLICATORS

/** cleanup is a function that takes a particles and prepares it
    for the next timestep */
class MDCleanup {
public:
  template< class P > 
  void operator()( P & p );
};

template< class P > void 
MDCleanup::operator()( P & p ){
  // reset force/torque value of particle
  p.f      = 0.;
  p.torque = 0.;
  //std::cout<<p.x<<std::endl;
  for ( typename P::CMapIterator i  = p.contact.begin(); 
	                         i != p.contact.end(); ) {

    if ( ! (*i).second.active ) { 
      // bond still present, but not ecnountered in 
      // in interaction routine, remove it
      p.contact.erase(i);
      PT_DEBUGINFO4( "removing inactive contact " << (*i).first );
    }
    else {
      (*i).second.active = false;
      ++i;
    }
  }

#if MOD_JANSEN_PRESSURE
  p.pressure = 0.;
#endif
}

#endif














