#ifndef MODPREDCORR5_H
#define MODPREDCORR5_H

#include "ModMaterial.H"

#ifndef MOD_SUPPRESS_Y_MOTION
# define MOD_SUPPRESS_Y_MOTION 0
#endif

#ifndef MOD_SUPPRESS_Z_MOTION
# define MOD_SUPPRESS_Z_MOTION 0
#endif

// -------------------------------------------
// Gear predictor-corrector
// -------------------------------------------

//namespace {  
  
// use the appropriate constants here - will be different for
// different order of the ODE (see Allen Tildesley - Computer
// Simulation of Liquids)
//const double gear0 = 19./120.;  // if x'' = f(x)
const double gear0 = 19./90.;     // if x'' = f(x,x') (vel proportional damp)
const double gear1 = 3./4.;
const double gear3 = 1./2.;
const double gear4 = 1./12.;

// we only integrate omega, not the orientation, thus 
// we use coefficients for the first order equation 
// (see Allen-Tildesley, S.341)
const double gearf0 = 251./720.;
const double gearf2 =  11./12.;
const double gearf3 =   1./3.;
const double gearf4 =   1./24.;
  
// -------------------------------------------
// predictor part
// -------------------------------------------

class MDGear5Pred {
protected:
  const double dt;             // time step
  const double c1, c2, c3, c4; // precalculated constants
public:
  // constructor for predictor
  MDGear5Pred( double a_dt): 
  dt(a_dt),        // store time step
  c1(dt),          // precalculate constans..
  c2(c1*dt/2.),    // ..
  c3(c2*dt/3.),    // ..
  c4(c3*dt/4.)     // ..
    {}
  
  // does predictor step for one particle (see below)
  template<class P> inline void operator()( P & p );
};
  
// does predictor step for one particle
template<class P> void MDGear5Pred::operator()( P& p )
{
  p.x  += (c1*p.v + c2*p.a + c3*p.x3 + c4*p.x4);
  p.v  += (         c1*p.a + c2*p.x3 + c3*p.x4);
  p.a  += (                  c1*p.x3 + c2*p.x4);
  p.x3 += (                            c1*p.x4);

  // show position
  //  cout << p.x[0] << " " << p.x[1] << " " << p.x[2] << endl;
} 

class MDRotGear5Pred : public MDGear5Pred {
public:
  MDRotGear5Pred( double a_dt ): MDGear5Pred(a_dt) {}
    
  template<class P> 
  void operator()( P& p ){
    MDGear5Pred::operator()(p);  // first do the translational part
    
    p.omega  += (c1*p.alpha + c2*p.omega2 + c3*p.omega3 + c4*p.omega4);
    p.alpha  += (             c1*p.omega2 + c2*p.omega3 + c3*p.omega4);
    p.omega2 += (                           c1*p.omega3 + c2*p.omega4);
    p.omega3 += (                                         c1*p.omega4);

  }
};


// -------------------------------------------
// corrector part
// -------------------------------------------

class MDGear5Corr {
protected:
  const Material material;
private:
  const double c0, c1, c3, c4; // precalculated corrector values
public:

  // !!!!!!!!! please check !!!!!!!!! - checked sts 3/2001
  // constructor for corrector
  MDGear5Corr( double a_dt, const Material &a_mat )
  : material(a_mat),  
    // c_n' = c_n * (n!/2) dt^(2-n)
    c0(gear0*a_dt*a_dt/2.),     // precalculate corr. constants
    c1(gear1*a_dt/2.),        // missing /2. in original file  
    c3(gear3*3./a_dt),        
    c4(gear4*12./(a_dt*a_dt))   
  { }
  
  // does corrector step for one particle (see below)
  template<class P> void operator()( P & );

};

// does corrector step for one particle
template<class P> void MDGear5Corr::operator()(P& p){
  // make sure that f finally stores an acceleration
  // in this way we can avoid requiring a mass member element here
  double_vec diff; 

#if MOD_SUPPRESS_Y_MOTION
  p.f[1] = 0.;
  PT_WARNING( "setting f[1] to " << p.f[1] << " on particle " << p.id );
#endif
#if MOD_SUPPRESS_Z_MOTION
  p.f[2] = 0.;
  PT_WARNING( "setting f[2] to " << p.f[2] << " on particle " << p.id );
#endif

  diff = p.f/material.getMass(p.type) - p.a;

  p.x  += c0 * diff;
  p.v  += c1 * diff;
  p.a  += diff;
  p.x3 += c3 * diff;
  p.x4 += c4 * diff;
} 

 
class MDRotGear5Corr : public MDGear5Corr {

  const double cf0, cf2, cf3, cf4;

public:

  MDRotGear5Corr( double a_dt, const Material &a_mat ) 
  : MDGear5Corr( a_dt, a_mat ),
    // c_n' = c_n * n! dt^(1-n)  (first order equation)
    cf0(gearf0*a_dt),
    cf2(gearf2* 2./a_dt),
    cf3(gearf3* 6./(a_dt*a_dt)),
    cf4(gearf4*24./(a_dt*a_dt*a_dt))
  { }

  template<class P>
  void 
  operator()( P& p ){

    MDGear5Corr::operator()( p ); // translation

    double_vec diff; 
    diff = p.torque/(material.getMomentOfInertia(p.type)) - p.alpha;

#if 0
    // testing: impose constant omega
    if ( p.id == 2  ){
      p.omega  = double_vec(0.,0.,4.);
      PT_WARNING( "imposing constant omega on id " << p.id << p.omega );
    }
#else
    p.omega  += cf0 * diff;
    p.alpha  += diff;
    p.omega2 += cf2 * diff;
    p.omega3 += cf3 * diff;
    p.omega4 += cf4 * diff;
#endif
  }   
};


#endif












