#ifndef MODWALL_H
#define MODWALL_H

#ifdef PTM_VISUALIZE
# include "XT/XTDisplay.h"
#endif 

#include "PT/PTZ_Matrix.h"
#include "PT/PTZ_Vector.h"
#include "PT/PT_Misc.h"
#include "PT/PTmath.h"
#include "PT/PTlimits.h"
#include "PT/PTS_StSum.h"

#include "ModParticle.H"    // defines PTS_CMapVector/MOD_PERSISTENT_CONTACTS


// ------------------------------------------------
// Wall feature: 
// abstraction for the idea of 'distance from a wall'
//  region of interest = region in which 'spheres' might be 'in contact'
//  with the wall, this is normally taken to be the geometrical specification 
//  of the wall rectangle(s) plus the particle radius - this might be too 
//  restrictive in cases where wall meet at angles slighlty smaller than 
//  180 dg.   
// ------------------------------------------------ 

// ------------------------------------------------
// currently supported
//   - Rectangle (infinite wall)
//   - ConvexFeature  2 half infinite walls joined by a common edge
//   - Cylinder (solid or hollow)
//   - Sphere   (solid or hollow)
//   - moving Rectangle 

struct WallFeature: public PTS_StSum<double> {

  WallFeature() : my_type(0), my_id(numeric_limits<unsigned long>::max()){}

  /** determines distance and direction of closest approach of a point, 
      the call always succeeds and returns continuous distances and 
      directions as required for the computation of a wall-sphere 
      interaction, roughness is a number between 0 and 1 which will be used 
      to modulate/multiply the regular friction constant for the 
      wall-particle contact */
  inline virtual void 
  overlap(const double_vec &p, double &dist, 
	  double_vec &normal, double_vec &velocity,
	  double &roughness)=0;

  /** extent of an axis parallel epiped (Quader) that is guaranteed 
      to contain all points in distance 'dist' from the feature */
  virtual void 
  region(double dist, double_vec &ll, double_vec &ur)=0;

  /** gives walls the chance to do something at the beginning 
      of each time step, base class increments configuration counter 
      in the statistics base class if t > 0
    */
  virtual void update(double t); 

  unsigned long get_id() 
  { 
    PT_ASSERT( my_id != numeric_limits<unsigned long>::max() );
    return my_id; 
  }

  unsigned get_type(){ return my_type; }

#ifdef PTM_VISUALIZE
  /** */
  virtual void 
  draw(XTDisplay &){}
#endif

  /** compute interaction, solely based on dist and normal from a 
      virtual call to overlap */
  template<class P>
  void operator()(P &p);
  template<class P> inline  
  double plastic_force(P & p, double overlap);
  template<class P> inline  
  double force_loading_Hertz( P & p, double overlap );
  template<class P> inline  
  double force_unloading_Hertz( P & p, double overlap,double xmax); 
  template<class P> inline  
  double decision( P & p, double overlap);
  template<class P> inline 
  double compute_xp(P & p,double xmax);
public:
  static Material material;
protected:
  friend class   Geometry;
  unsigned long  set_id(unsigned long id) { return my_id=id; }

  unsigned       my_type; // material type, default =0
  unsigned long  my_id;   // id for contact lists, default=0
                          //  set explicitly in main, processor global
};

// =================================================================
// INTERACTION COMPUTATION MOVED TO ModForces.H
// because of the strong similarity between particle-particle and
// particle-wall interaction, it is better to have both routines in 
// one file; the likelihood to forget updating one of the routines 
// is smaller then. 
// =================================================================
// template<class P>
// inline void
// WallFeature::operator()(P &p) ...


// -------------------------------------------------------------
// Class Rectangle
// -------------------------------------------------------------

struct Rectangle: public WallFeature {
  Rectangle(){}
  Rectangle(istream &);

  // pure virtuals to be implemented
  inline void overlap(const double_vec &p, double &dist, 
		      double_vec &normal, double_vec &vel, double &roughness);
  void region(double dist, double_vec &ll, double_vec &ur);

  // a function to determine whether the projected normal to the point 
  // intersects the rectangle
  bool is_inside( const double_vec &p );
#ifdef PTM_VISUALIZE
  void draw(XTDisplay &window);
#endif

  double_vec get_normal() const { return m_normal; }
  double_vec get_vertex(int i) const;

  friend std::istream &
  operator>>(std::istream &in, Rectangle &r);

public:
  static void 
  set_show_normal( bool new_state ) { our_show_normal = new_state; }

protected:
  double m_d;      // dot(normal,ref_point)
  double_vec m_normal; // normal
  double_vec m_points[4];
  double m_color;  
  static bool our_show_normal;
};


inline void 
Rectangle::overlap(const double_vec &p, double &dist, 
		   double_vec &normal, double_vec &vel, double &roughness )
{
  (void) roughness; (void) vel;  // keep gcc froom complaining

  dist   = dot(p,m_normal) - m_d;
  // PT_DEBUGINFO( p << m_normal << m_d << " dist: " << dist);
  normal = m_normal;
}

ostream &
operator<<(ostream &, const Rectangle &);

// -----------------------------------------------------------------
// Convex - a 'series' of wall rectangles exposing angles < 180 degrees 
//  to the interior of the particle container, these are treated as 
//  one 'feature' 
// -----------------------------------------------------------------


/* a correct and complete treatment of walls works probably as follows
   (I) segment the walls into nonoverlapping polygons (these can be 
       further reduced to triangles if that is convenient)
   (ii) for each polygon, decide whether it interacts with a particle or 
       not. The decision is solely based on distance from the closest point 
       of the polygon. It helps to associate with each side of the polygon 
       a normal lying in the polygonal plane, pointing to its interior. Then 
       the sign of (point-ref).normal - ref.normal 
       indicates whether point is 'above/inside' the polygon or not. 
       If the answer is above/inside for all sides of the polygon, 
       compute the distance from the polygon surface in the same fashion
       (point-ref).normal - ref.normal and mark the polygon 
       as 'interacting' with the particle
   (iii) now consider all edges of polygons. If the edge constitutes an angle 
       greater than 180 dg. (seen from the interior) 
       we must take special care, because the particle 
       might not interact with any one of the two adjacent polygons. If this 
       is the case (check the 'interacting' flags), we consider 
       whether the perpendicular projection yields a point on the line 
       (inside/interacting) or beyond the end points (outside). we mark
       the edge to be 'interacting' and compute the distance if the 
       first is the case, otherwise there is no interaction with the edge. 
       (we would also mark the edge as interacting, if any of the adjacent
       polygons 'interact')
   (iv) we last run through all nodes connecting edges. If none of the 
       edges joining at the node is found 'interacting', we must still 
       consider the particle-node distance. Only if this distance is larger 
       than the particle radius then there is no particle-wall interaction 
       for this particular particle
 */
/* 
   We are going to simplify the procedure significantly for our purposes:
   we assume that the wall can be decomposed into rectangles of which at 
   most two form edges with angles > 180 dg. As a second form of wall 
   we use single rectangles which are in fact treated as
   just representing entire planes. (These two features will still 
   allow (by 'intersection') for trapezoidal forms as required e.g. at 
   constriction points)

   The interaction is assumed to be additive with respect to these two forms
   of 'wall features'. Care must be taken to guarantee smooth interactions 
   at angles close to 180 where the sphere might interact with more than
   one 'feature' - here it must be guaranteed that the 'range' of a feature
   (i.e. the geometric criterion used to decide whether we look at 
   a possible sphere-feature interaction in the 
   first place) is large enough.
 */    

struct ConvexFeature : public WallFeature {

  ConvexFeature(istream &);

  void overlap(const double_vec &p, double &dist, 
	       double_vec &normal, double_vec &vel, double &roughness);
  void region(double dist, double_vec &ll, double_vec &ur);

#ifdef PTM_VISUALIZE
  void draw(XTDisplay &);
#endif
private:
  // vertex points of the edge (=common corners of the rectangles ):
  // I need the PTZ_vector here, because they provide ctors for istream 
  PTZ_Vector<double_vec,2> vertex; 
  PTZ_Vector<Rectangle,2>  rect; // adjacent rectangles
  PTZ_Vector<double_vec,2> edge_normal; // edge normals in the rectangle planes
  double_vec               edge_dir; // normalized edge vector 0 -> 1 
  double color;
};

inline void 
ConvexFeature::overlap(const double_vec &p, double &dist, 
 	           double_vec &normal, double_vec &vel, double &roughness )
{
  (void) roughness; (void) vel; 

  // check whether the particle 'overlaps' with one of the rectangles, 
  // if not, compute the distance from the edge
  // if you don't touch roughness, it will be taken to equal one.

  // most common case first
  double in_plane = dot(edge_normal[0],p-vertex[0]);
  if ( in_plane > 0 ) { // interaction with 'upper plane' 
    rect[0].overlap(p,dist,normal,vel,roughness);
    return; 
  }
  else { 

    in_plane = dot(edge_normal[1],p-vertex[0]);
    if ( in_plane > 0 ) { // interaction with 'lower plane'
      rect[1].overlap(p,dist,normal,vel,roughness);
      return;
    }

    // if we are here, we need to compute the distance to the edge
    // we do not change roughness
    double_vec d = p - vertex[0];
    d -= dot(edge_dir,d)*edge_dir;  // perpendicular component of p-vertex[0]
    normal = d; 
    dist = sqrt( dot(d,d) );
    normal /= dist;
    return;
  }
}

// -------------------------------------------------------------------
// CYLINDER - just a cylindrical wall specified by its center axis and 
//            radius
// -------------------------------------------------------------------

struct Cylinder: public WallFeature {

  Cylinder(istream &);

  void overlap(const double_vec &p, double &dist, 
	       double_vec &normal, double_vec &vel, 
	       double &roughness);

  void region(double dist, double_vec &ll, double_vec &ur);

#ifdef PTM_VISUALIZE
  void draw(XTDisplay &);
#endif  
private:
  double radius;          // positive = hollow cylinder, negative = solid
  double_vec bottom_point;
  double_vec top_point;
  double_vec axis;        // normalized direction of cylinder
};


inline void
Cylinder::overlap(const double_vec &p, double &dist, 
		  double_vec &normal, double_vec &vel, double &roughness )
{
  (void) roughness; (void) vel;

  double_vec p_rel  = p - bottom_point;
  double_vec p_perp = p_rel - dot(p_rel,axis)*axis; // axis -> point
  double     l_perp = sqrt( dot(p_perp,p_perp) );   // distance to axis, > 0

  if ( radius >= 0 ){
    dist            = radius - l_perp;
    // normal will be directed outward for negative radii (solid cylinders)
    normal          = -p_perp / l_perp;
  }
  else {
    dist            = l_perp + radius;
    normal          = p_perp / l_perp;
  }

  // no roughness modeled

}



// -----------------------------------------------------------
//  SPHERE  cavity or obstacle
// -----------------------------------------------------------


struct WallSphere: public WallFeature {

  WallSphere(istream &);

  void overlap(const double_vec &p, double &dist, 
	       double_vec &normal, double_vec &vel, double &roughness);
  void region(double dist, double_vec &ll, double_vec &ur);

#ifdef PTM_VISUALIZE
  void draw(XTDisplay &);
#endif  
private:
  double radius;          // negative = hollow, positive = solid
  double_vec center;
};


inline void 
WallSphere::overlap(const double_vec &p, double &dist,
		    double_vec &normal, double_vec &vel, double &roughness )
{
  (void) roughness; (void) vel;

  normal = p - center;
  dist   = sqrt( dot(normal,normal) );
  if ( radius >=0 ){  // positive radius implies solid sphere
    normal /= dist;
    dist -= radius;
  }
  else {
    normal /= -dist;  // negative radius: hollow sphere
    dist = -radius-dist;
  }
}

// -----------------------------------------------------------------------
// 'rectangular' wall moving along the surface normal 
// -----------------------------------------------------------------------

struct MovingRectangle: public Rectangle {

  MovingRectangle(istream &);

  inline void overlap(const double_vec &p, double &dist, 
	       double_vec &normal, double_vec &vel, double &roughness);
  void region(double dist, double_vec &ll, double_vec &ur);

  // displacement along normal direction
  void update(double t);

#ifdef PTM_VISUALIZE
  void draw(XTDisplay &);
#endif  
private:
  double m_dplmt;               // displacement along surface normal
  double m_begin, m_end, m_vel; // start/end motion, linear velocity
};


inline void 
MovingRectangle::overlap(const double_vec &p, double &dist, 
		         double_vec &normal, double_vec &vel, 
			 double &roughness )
{
  (void) roughness; 

  dist   = dot(p,m_normal) - m_d - m_dplmt;
  // PT_DEBUGINFO( p << m_normal << m_d << " dist: " << dist);
  normal = m_normal;
  vel    = m_normal * m_vel;   // velocity of the surface
}


#endif













