
#include "ModWall.H"

Material WallFeature::material;

void WallFeature::update(double t){
  // indicate 'previous timestep is over' to the statistics class
  if ( t > 0. ) 
    next_config();
}


// ---------------------- RECTANGLE ---------------------

// read rectangle from geometry file (ctor called by PTS_Factory in 
// Geometry::Geometry)
Rectangle::Rectangle(istream &in)
{
  in >> *this >> my_type;
  PT_ASSERT( (int)my_type < material.size() );
}


std::istream &
operator>>(std::istream &in, Rectangle &r){
  // reads 4 double_vecs from a stream and constructs a rectangle from them 
  // dont read the material to support simpler input format in ConvexFeature

  PT_VERIFY( in );
  in >> r.m_points[0];
  in >> r.m_points[1];
  in >> r.m_points[2];
  in >> r.m_points[3];
  
  // p1-p0,p3-p0,m_normal follow r.h. rule
  r.m_normal =  cross(r.m_points[1]-r.m_points[0],r.m_points[3]-r.m_points[0]); 
  r.m_normal /= sqrt( dot(r.m_normal,r.m_normal) );
  r.m_d      =  dot(r.m_normal,r.m_points[0]); 

  return in;
}
  
std::ostream &
operator<<(std::ostream &o, const Rectangle &r){
  for ( int i=0; i < 4; ++i )
    o << r.get_vertex(i);
  return o;
}

void 
Rectangle::region(double dist, double_vec &ll, double_vec &ur){
  ll = std::numeric_limits<double>::max();
  ur = std::numeric_limits<double>::min();

  for ( int dim=0; dim < ll.size(); ++dim ){
    for ( int i=0; i < 4; ++i ){
      ll[dim] = min(ll[dim],m_points[i][dim]-dist);
      ur[dim] = max(ur[dim],m_points[i][dim]+dist);
    }
  }
}

double_vec 
Rectangle::get_vertex(int i) const {
  PT_ASSERT(i<4);
  PT_ASSERT(i>=0);
  return m_points[i];
}

// static member
bool Rectangle::our_show_normal = false;

#if 0  
// unused for now - could help if we have to deal with many wall segments
// and very complex protrusion, indentation gemoetries - untested       
bool
Rectangle::is_inside(const double_vec &p){

  // compute the components of p - p0 with respect to the system spanned 
  // by p1-p0, p3-p0, m_normal. By construction, these three vectors should 
  // form a orthogonal system
  // 
  // I'll go for the safe option, if I invert the matrix containing
  // the base vectors as columns and apply that matrix to a vector, 
  // then I get the components as result. 

  PTZ_Matrix<double,3,3> inv_base;
  for ( int j=0; j < 3; ++j ){
    inv_base[j][0] = m_points[1][j]-m_points[0][j]; 
    inv_base[j][1] = m_points[3][j]-m_points[0][j];   
    inv_base[j][2] = m_normal[j];   
  }
  inv_base.invert(); 
  
  // now apply to p-p0 and look at the first two components of the result
  PTZ_Vector<double,3> res = inv_base * (p-m_points[0]);
  if ( 0 <= res[0] && res[0] <= 1 && 0<= res[1] && res[1] <= 1 )
    return true; 
  else 
    return false;
}
#endif


// ------------------------ CONVEX FEATURE

ConvexFeature::ConvexFeature( istream &in ) 
{
  PT_ASSERT( in );
  in >> vertex >> rect;  // vectors themselves

  // we have been called because a key word has been found in the file, 
  // it is left to us to parse the remaining data. 
  // for future generalizations, we should expect a vector/list of 
  // adjacent wall segments.
  // for now the format in the input file is 
  // convex [ vertex0 vertex1 ] [ Rectangle0 Rectangle1 ] color
  // where vertex0 is the input format expected by double_vec and
  // Rectangle the format expected by Rectangle 
  // the angle between the normals of the Rectangles is expetec to be 
  // 0 < angle < 180
  // vertex0 and vertex1 are not expected to be common points with the 
  // two rectangles, but to be located on their intersection line 

  in >> my_type;  // material for the feature
  if ( (int) my_type >= material.size() ) 
    PT_SEVERE( "convex feature mat. > no. materials" ); 
  PT_VERIFY( in.good() );

  // compute a normal to the common edge lying in the rectangle plane 
  // and guarantee that it points to the interior of the rectangle 
  edge_dir  = vertex[1] - vertex[0];  
  edge_dir /= sqrt( dot(edge_dir,edge_dir) );

  for ( int rect_id=0; rect_id < 2; ++rect_id ){

    // first attempt, could be wrong by a sign
    edge_normal[rect_id] = cross(rect[rect_id].get_normal(),edge_dir);
    
    // run over the rectangle vertices and check that they yield all 
    // 'positive' distances from the edge; otherwise flip the normal
    double dist = 0; 
    for ( int rect_vert=0; rect_vert< 4; ++rect_vert) {
      double_vec vert_cur = rect[rect_id].get_vertex(rect_vert) - vertex[0]; 
      dist = dot( vert_cur, edge_normal[rect_id] );
      if ( dist < 0 ) break;
    }
    if ( dist >= 0 ) { 
      break; // all rectangle points yield poostive values 
    }
    else { // try the other direction 
      edge_normal[rect_id] *= -1.;
      // test for consistent input data:
      for ( int rect_vert=0; rect_vert< 4; ++rect_vert) {
	double_vec vert_cur = rect[rect_id].get_vertex(rect_vert) - vertex[0]; 
	dist = dot( vert_cur, edge_normal[rect_id] );
	if ( dist < 0 ){ 
	  PT_SEVERE( "inconsistent input data for ConvexFeature ctor,"
		     " cannot determine edge normal" );
	  PT_ASSERT(0);
	}
      } // for
    } // else (dist < 0)
  } // for rect_id
}


void 
ConvexFeature::region(double dist, double_vec &ll, double_vec &ur){
  // take the 'union' of the rectangles and the edge, taking 
  // dist into account, and return
  ll = std::numeric_limits<double>::max();
  ur = std::numeric_limits<double>::min();

  double_vec r1ll, r1ur;  // rectangle 1 lower left/upper right
  rect[0].region(dist,ll,ur);
  rect[1].region(dist,r1ll,r1ur);

  for ( int dim=0; dim < ll.size(); ++dim ){
    for ( int i=0; i < 2; ++i ){
      ll[dim] = min(ll[dim],vertex[i][dim]-dist);
      ur[dim] = max(ur[dim],vertex[i][dim]+dist);
    }
  } 
} 

// -------------- CYLINDER --------------------------------------

Cylinder::Cylinder( istream & in ){

  in >> bottom_point >> top_point >> radius >> my_type;
  PT_SOMECHECKS( 
  if ( in.fail() ) 
    PT_WARNING ( "cylinder specification incorrect" );
  if ( (int) my_type >= material.size() ) 
    PT_SEVERE( "cylinder mat. >= no. materials" ); 
  )

  axis = top_point - bottom_point;
  axis /= sqrt( dot(axis,axis) );
}


void
Cylinder::region(double dist, double_vec &ll, double_vec &ur)
{
  ll = std::numeric_limits<double>::max();
  ur = std::numeric_limits<double>::min();

  PT_ASSERT(radius > 0); // needs thinking for radius < 0
  for ( int dim = 0; dim < ll.size(); ++dim ){
    ll[dim] = min(ll[dim],bottom_point[dim]-radius-dist);
    ll[dim] = min(ll[dim],top_point[dim]-radius-dist);
    ur[dim] = max(ur[dim],bottom_point[dim]+radius+dist);    
    ur[dim] = max(ur[dim],top_point[dim]+radius+dist);
  }
}

// ----------------- sphere -------------------------------------

WallSphere::WallSphere( istream &in )
{
  in >> center >> radius >> my_type;
  if ( (int) my_type >= material.size() ) 
    PT_SEVERE( "wall sphere mat. >= no. materials" ); 
}

void 
WallSphere::region(double dist, double_vec &ll, double_vec &ur)
{
  ll = std::numeric_limits<double>::max();
  ur = std::numeric_limits<double>::min();

  PT_ASSERT(radius > 0); // think ...
  for ( int dim = 0; dim < ll.size(); ++dim ){
    ll[dim] = min(ll[dim],center[dim]-radius-dist);
    ur[dim] = max(ur[dim],center[dim]+radius+dist);    
  }
}

// ----------------- moving rectangle ---------------------------

MovingRectangle::MovingRectangle( istream &in )
: Rectangle(in), m_dplmt(0.), m_begin(0.), m_end(0.), m_vel(0.) 
{
  in >> m_begin >> m_end >> m_vel;
  if ( m_end < m_begin ) 
    PT_SEVERE( "MovingRectangle: start > stop " 
	       << m_begin << " > " << m_end );
}

void MovingRectangle::region(double dist, double_vec &ll, double_vec &ur)
{
  ll = std::numeric_limits<double>::max();
  ur = std::numeric_limits<double>::min();

  for ( int dim=0; dim < ll.size(); ++dim ){
    for ( int i=0; i < 4; ++i ){
      ll[dim] = min(ll[dim],m_points[i][dim] + m_dplmt*m_normal[dim] - dist);
      ur[dim] = max(ur[dim],m_points[i][dim] + m_dplmt*m_normal[dim] + dist);
    }
  }
}

void MovingRectangle::update(double t)
{
  WallFeature::update(t);  // takes care of the statistics class

  double diff = t - m_begin;
  if ( diff > 0 && t < m_end ) 
    m_dplmt = m_vel * diff; 
  else 
    return;
}


// ----------------- DRAW member functions ----------------------

#ifdef PTM_VISUALIZE
void 
Rectangle::draw(XTDisplay &window){

  XTPoint list[4];
  XTPoint center(0); 

  for (int i=0; i<4; i++){
    list[i]=XTPoint(m_points[i][0],m_points[i][1],m_points[i][2]);
    center += 0.25 * list[i];  // center of mass
  }
      
  window.set_foreground_color( material.getColor(my_type) );
  window << XTPolygon(4, list);
  // draw the normal so that we can be sure fo the orientation of the walls
  if ( our_show_normal ) {
    XTPoint normal(m_normal[0],m_normal[1],m_normal[2]);
    window.line( center, center+normal );
  } 
}

void 
ConvexFeature::draw(XTDisplay &window){
  
  window << XTColor( material.getColor(my_type) );
  rect[0].draw(window);
  rect[1].draw(window);

  XTPoint  v0(vertex[0][0],vertex[0][1], vertex[0][2] );
  XTPoint  v1(vertex[1][0],vertex[1][1], vertex[1][2] );
  window.line( v0, v1 );

  // draw the edge normals - should point to the inner of the respective 
  // rectangles
  PT_DEBUG(
  XTPoint  en0(edge_normal[0][0],edge_normal[0][1], edge_normal[0][2] );
  XTPoint  en1(edge_normal[1][0],edge_normal[1][1], edge_normal[1][2] );
  window.line( .5*(v0 + v1),.5*(v0 + v1)+ en0 );   	   
  window.line( .5*(v0 + v1),.5*(v0 + v1)+ en1 );   	   
  )
}

void 
Cylinder::draw(XTDisplay &window){
  // fake a cylinder using two circles
  window << XTColor( material.getColor(my_type) );
  double_vec diff = top_point - bottom_point;
  XTPoint    ori   = XTPoint(diff[0],diff[1],diff[2]);
  window << XTCircle( 
	      XTPoint(bottom_point[0],bottom_point[1],bottom_point[2]), 
	      radius, ori );
  window << XTCircle(
	      XTPoint(top_point[0],top_point[1],top_point[2]), 
	      radius, ori );
}

void
WallSphere::draw(XTDisplay &w)
{
  w << XTColor( material.getColor(my_type) );
  w << XTSphere(XTPoint(center[0],center[1],center[2]),radius);
}

void 
MovingRectangle::draw(XTDisplay &window){

  XTPoint list[4];
  XTPoint center(0); 

  for (int i=0; i<4; i++){
    double_vec vertex = m_points[i] + m_dplmt * m_normal; 
    list[i] = XTPoint(vertex[0],vertex[1],vertex[2]);
    center += 0.25 * list[i];  // center of mass
  }
      
  window.set_foreground_color( material.getColor(my_type) );
  window << XTPolygon(4, list);
  // draw the normal so that we can be sure of the orientation of the walls
  if ( our_show_normal ) {
    XTPoint normal(m_normal[0],m_normal[1],m_normal[2]);
    window.line( center, center+normal );
  } 
}


#endif







