#include <time.h>
#include <ctype.h>

#include <string>
#include <iostream>
#include <fstream>
#include <memory>

#include "PT/PTC_Para.h"
#include "PT/PTC_Log.h"
#include "PT/PTM_PartContLC.h"
#include "PT/PTM_ParPartCont.h"
#include "PT/PTM_MDAlgorithm.h"
#include "PT/PTSRandom.h"
#include "PT/PTSSimClock.h"
#include "PT/PTSSimFile.h"
#include "PT/PTSTimer.h"
#include "PT/PTSSignal.h"

#if PT_PARALLEL
# include "PT/PTS_MPIComm.h"
#else 
# include "PT/PTS_SerComm.h"
  typedef PTS_SerCommEnv PTS_MPICommEnv;
#endif


#include "ModParticle.H"  // typedef for double/int_vec 

#ifdef PTM_VISUALIZE
# include "XT/XTDisplay.h"
# include "XT/XTGLDev.h"
#endif 

#include "ModPredCorr5.H"  
#include "ModParticle.H"  
#include "ModForce.H"  
#include "ModGeometry.H"  
#include "ModGravity.H"  
#include "ModControl.H"  // defines ParticleBox
#include "ModIO.H"       // save/restart/measure 

#include "PT/PTDebug.h"


int main(int argc, char **argv)
{  
  PTC_SysInfo  si;
  PT_INFO( "starting on " << si.host() << " in dir " << si.dir() ); 

  // initialize MPI in a somewhat fancy way...
  PTS_MPICommEnv mpi(argc,argv);

  // you can specify 'pfile=<filename>' in the command line, otherwise 
  // <progname>.cnf is expected to be present, 'wd=<directory>' specifies a 
  // working directory
  PTC_Para pf;          
  if ( mpi.node_id() == 0 ){
    pf.standard_init(argc, const_cast<const char **>(argv)); 
  }
  mpi.broadcast( pf ); // should now be the same everywhere.
  if ( pf.exists("wd") ){
    chdir( pf.getd("wd", "") );
  }
  // in a multi proc environment, I inject node_str so that the restart files
  // can be written on per-node basis (used by PTS_SimFile). 
  pf.set("node_str",PTi2string(mpi.node_id(),2));
  // this was used by Control, but may be somewhere else. 
  // Since it does not harm, I leave it for the moment
  pf.set("node_id",mpi.node_id());

  // write parameters to protocol file
  pf.print_state(cerr);
  
  string strProject = pf.getd("project_name","");  
  // Project name specified?
  if (strProject==""){
    std::cerr << "No project_name specified in configuration file" << endl;
    return (-1);
  } 

  // collect I/O on a per CPU basis in logfile
  PTC_LogBuffer::log_mode  lm = PTC_LogBuffer::silent;
  if ( pf.getd("use_logfile_and_cout",false) )
    lm = PTC_LogBuffer::cp2stdout;
  PTC_Log logfile((strProject+PTi2string(mpi.node_id(),2)+".log").c_str(),lm);
  //PTS_Redirect redir_cout(cout, logfile);
  //PTS_Redirect redir_cerr(cerr, logfile);

  // print parameter settings
  pf.print_state(logfile);       

  // read parameters or take defaults 
  double max_t   = pf.getd("runtime",10.);      //total sim. time
  double max_walltime = pf.getd("max_walltime",3.8*3600.); // batch time
  int_vec cpuvec = pf.getd("cpus",int_vec(1,1,1)); //number of cpus in each direction
  int_vec bc     = pf.getd("bound_cdt",int_vec(0));  //boundary condition

  // ------  material parameters from materials file:
  Material material(pf);

  // this isn't very clean, but one of the ways to make sure that the 
  // compiler can optimize the 'material'.get calls in the wall interaction
  // computation is to make material a static member of the WallFeatures. 
  // As we have to read the material properties frm a file, the initial 
  // static initialization must be overridden with the correct values
  WallFeature::material = material;

  // ------  some 'counters', providing intervals for periodic activities 
  PTS_SimClock               graphics("graphics",pf);
  PTS_SimFile<PTS_SimClock>  save("save",pf,strProject);
  // we will check for walltime on node 0, then broadcast it to get the 
  // time to the other nodes in consistent fashion
  PTS_Timer<PTS_WallTime>    wall_clock; wall_clock.start();
  PTS_SimFile<PTS_SimClock>  restart("restart",pf,strProject);
  PTS_SimClock               ld_balance("ld_balance",pf);
  PTS_ResClock<PTS_WallTime> progress_info("progress_info",pf);

  PTS_CpuTime                cputime;
  PTS_Signal<SIGUSR1>        sig_usr1;   // do something when USR1 is raised
//#if PT_CRAY_T3E
//  PTS_Signal<SIGFPE>         sig_fpe; 
//#endif

  // now pick a container and cell size that guarantees that we will see all 
  // pair interactions
  // if the particle radius = 1 and each cell contains the particle 
  // that have their CM inside, then the dimension of the cell must be 
  // larger than 2
  double cutoff = material.get_cutoff();
  PT_INFO( "cutoff is " << cutoff );
  double dt     = material.get_timestep_gpk5();   // timestep
  PT_INFO( "timestep is " << dt );

  // setup
  //PTS_KPS250RNG r;         // random number generator:
  //PTS_GaussDist gauss(r);  // gauss distributed numbers

  // event specifications
  Control control(material,pf,mpi);

  // geometry / wall specification
  Geometry  geometry(pf);
  double_vec ll=geometry.ll;
  double_vec ur=geometry.ur;
  geometry.set_wallfeature_ids(control,mpi); // globally unique wall ids
  
#ifdef PTM_VISUALIZE

  std::auto_ptr<XTDevice> auto_w;
  std::string  graphics_device = pf.getd("graphics_device","OpenGL");
  if ( graphics_device == "OpenGL" ) {
    // ------ visualization stuff:
    //XTX11Dev w(500,500,argc,argv); // X11 2d graphics
    //XTPovDev w("out-film");        // Pov Ray rendering
    std::auto_ptr<XTDevice> tmp(new XTGLDev(500,500,argc,argv)); 
    auto_w = tmp; 
  }
  // else if ( graphics_device == "Null" )

  XTDisplay  window(auto_w.get());                
  XTColormap map(XTColormap::rainbow);

  if ( graphics_device != "Null" ) {
    window.set_colormap_mode(XTDevice::indexed);
    window.set_colormap(map);
    // user color range 
    XTCrange col_range(0,10);
    window.set_color_range(col_range);
    window.set_background_color(XTWhite);
    window.set_foreground_color(XTBlack);
    
    // coordinate range, perspective 
    double_vec center(0.5*(ur+ll)), span(ur-ll);
    // its  essential that the eye position is outside the coordiante range, 
    // even as we rotate
    double     max_span = PTmax(span[0],span[1],span[2]);
    XTPoint    eye_pos(center[0]+1.5*max_span,center[1]-3*max_span,center[2]);
    XTPoint    look_at(center[0],center[1],center[2]);
    XTPoint    up_dir(0,0,1); 
    XTObserver xo(eye_pos,look_at,up_dir);
    XTViewport vp;
    vp.set_pc(XTPoint(ll[0]-0.3*span[0],ll[1]-0.3*span[1],ll[2]-0.3*span[2]),
	      XTPoint(ur[0]+0.3*span[0],ur[1]+0.3*span[1],ur[2]+0.3*span[2]));
    vp.set_observer(xo);
    window  << vp; // set viewport
  }

  // do we want to see normals displayed with rectangles?
  Rectangle::set_show_normal( pf.getd("use_rectangles_with_normals",false) );

#endif  

  // initialize cartesian topology in MPI
  PTM_Parallel par_info(argc,argv,DIM,&cpuvec[0],&bc[0]);

  // -- and create the empty particle container 
  ParticleBox  box(ll,ur,bc,cutoff,par_info);

  // select walls active in our coordinate range, cutoff is a measure of
  // of the 'range'/vicinity that needs to be taken into account, if 
  // in doubt specify something large compared to the particle radius
  geometry.select_features(cutoff,box.local_ll(),box.local_ur());

  logfile << "box domain: " << ll << "/ " << ur << endl;;

  // ----------------------
  // SETUP CONFIGURATION
  // ----------------------

  double t=0.;          // simulation time  
  int t_step = 0;       // disc. number of steps
  double walltime = 0.; // correct time will be set for all procs 
                        //  in the simulation loop, necessary for 
                        //  syncing parallel operations

  // shall we restart if we find a restart file?
  if ( pf.getd("use_restart",false) ){

    logfile << "restart file: " << restart.fname() << std::endl;
    ifstream restart_in ( restart.fname().c_str() );
    // try to load it (expect return value 0):
    PT_FALSIFY( ReadRestartFile( restart_in, box, geometry, 
				 control, t, t_step, pf) );

  } 

  // write box size to file
  logfile << "box.size() = " << box.size() << " object(s)" << endl;
  logfile << "geometry.size() = " << geometry.size() 
	  << " wall feature(s)" << endl;

  // ---------------------------------------------------
  // start simulation
  // ---------------------------------------------------

  PT_DEBUGINFO( box );
  logfile << "starting simulation" << std::endl;

  // forces
  ForceModel1  force(material);   // hertzian force model
  MDCleanup    cleanup;           // sets all forces/torques to 0
  MDGravity    grav(material);    // sets gravity

  // integrators (gear 5)
  MDRotGear5Pred  int_pred5(dt);
  MDRotGear5Corr  int_corr5(dt,material);

  // main loop

  while( t < max_t && walltime < max_walltime ){

    PT_DEBUGINFO4( "time step " << t_step << ", t= " << t );
    control.processEvents(t, box);

    // reset force vector of all spheres, remove inactive bonds
    for_each_particle(box.begin(),box.end(),cleanup);
    
    // do predictor step for each sphere
    for_each_particle(box.begin(),box.end(),int_pred5);
    
    // sort particles into linked cell structure and start communication 
    // every other time step perform load balancing - update must be called 
    // after load balancing which only shifts processor boundaries
    if ( ld_balance.tick(t_step) ){
      box.load_balance(DIM-1); // shift boundaries in the z direction
      // select walls according to local coordinate ranges
      geometry.select_features(cutoff,box.local_ll(),box.local_ur());
    }
    box.update();   
    // give 'moving' walls the chance to adjust before particle-wall int.
    geometry.update(t); 
    
    // set gravity for all spheres
    for_each_particle(box.begin(),box.end(),grav);
    
    // particle-particle interaction
    box.for_each_pair(force);

    // wall particle interaction
    for_each_particle(box.begin(),box.end(),geometry);
    // --------------------------------------------

    // do corrector step for each sphere
    for_each_particle(box.begin(), box.end(),int_corr5);
    
    //advance time:
    t += dt;
    t_step++;
    
    //drawing: (if enabled)
#ifdef PTM_VISUALIZE
    // clear window
    if ( graphics.tick(t) && graphics_device != "Null" ){
      window.clear();
      
      // draw all spheres
      for ( ParticleBox::iterator i=box.begin(); 
	   i != box.all_end(); 
	   ++i ) {
	XTPoint  center( (*i).x[0],(*i).x[1],(*i).x[2]);
	window.set_foreground_color( material.getColor((*i).type) );
	window << XTSphere( center, material.getRadius((*i).type) );
      }
      
      // draw walls
      geometry.draw(window);
    
      // draw simulation box
      window.set_foreground_color(XTBlack);
      double_vec ur = box.ur();
      double_vec ll = box.ll();
      window.rectangle(XTRectangle(XTPoint(ll[0],ll[1],ll[2]),
				   XTPoint(ur[0],ur[1],ur[2])));
    
      // ... label the axes
      window << XTText(XTPoint(ur[0],ll[1],ll[2]),"x");
      window << XTText(XTPoint(ll[0],ur[1],ll[2]),"y");
      window << XTText(XTPoint(ll[0],ll[1],ur[2]),"z");

      // ... and show our piece of the pie (will differ in the parallel case) 
#if PT_PARALLEL > 0
      window.set_colormap_mode( XTDevice::rgb );
      window << XTColor( 0.1, 0.9, 0.1 );
      ur = box.local_ur();
      ll = box.local_ll();
      window.rectangle(XTRectangle(XTPoint(ll[0],ll[1],ll[2]),
				   XTPoint(ur[0],ur[1],ur[2])));
      window.set_colormap_mode( XTDevice::indexed );
      window << XTBlack;
#endif 

      // show everything
      window.display();
      ProgressInfo(logfile, box, t, t_step, cputime);
    }
#endif

    // save output file
    if ( save.tick(t) ){
      // for now in ModControl.cc
      logfile << "writing data file..."; 
      WriteSaveFile( save.stream(), t, box, geometry, mpi, pf );
      logfile << "done." << std::endl; 
    }

    // it must be absolutely certain that the restart files are
    // created at the exact same deterministic spot in all instances
    // of the program. Thus, we query for cpu time on node 0,
    // broadcast it and then perform the test whether or not to write
    // a restart file we can take other routes (barrier), if it turns
    // out that the broadcast is too expensive wrt. to the other
    // communication in the program (which I doubt).
    if ( mpi.node_id() == 0 ){
      walltime = wall_clock.elapsed();
    }
    mpi.broadcast(walltime,0);

    // save temporary restartfile(s)
    if ( restart.tick(walltime) ){
      logfile << "writing restart file...";
      WriteRestartFile(restart.stream(), box, geometry, control, 
		       t, t_step, pf);
      logfile << "done." << std::endl;
    }

    if ( progress_info.tick() ){
      // unsynchronized stuff (do not communicate in here)
      ProgressInfo(logfile, box, t, t_step, cputime);
      logfile << "pending events: " << control.active_events() << std::endl;
    }

    if ( sig_usr1.exceeded() && mpi.node_id() == 0 ) {
      ProgressInfo(std::cout, box, t, t_step, cputime );

#if PT_PARALLEL == 0  // interactive stuff in the single processor version  
      bool done = false;
      while ( !done ) {
	cout << "(c) continue\n"
#if PTM_VISUALIZE
	        "(s) snapshot (to `snapshot.gif')\n"
#endif
	        "(x) exit NOW (no restart file written)\n ";
	char c;
	do { 
	  char in[5];
	  cin.getline(in,5);
	  c = in[0];
	} while ( isspace(c) );
	
	done = true;
	switch ( c ) {
	case 'c' : cout << "continuing." << std::endl; break;
	case 'x' : cout << "unconditional exit\n"; exit(1); break;
#if PTM_VISUALIZE
	case 's' : 
	  if ( dynamic_cast<XTX11EvWin*>(&(*auto_w)) == 0 ) 
	    cout << "device not capable of snapshots\n"; 
	  else {
	    dynamic_cast<XTX11EvWin*>(&(*auto_w))->save_file( "snapshot.gif" );
	    cout << "`snapshot.gif' created." << endl;
	  }
	  done = false;
	  break;
#endif
        // this must be synchronized over CPU's to avoid hanging and to 
        // provide consistent information	
        //case 'r' : WriteRestartFile(restart.stream(),box,control,t,t_step);
        //	 break;
	default  : done = false; break;
	}
      }
#endif // interactive stuff

      sig_usr1.set_limit( sig_usr1.now() + 0.5 );
    }


  } // while t < t_max && wall_time < max_walltime
    
  ProgressInfo(logfile, box, t, t_step, cputime);

  // always create restart file at the end 
  logfile << "simulation ended normally after " << walltime 
	  << "s" << std::endl;
  WriteRestartFile(restart.stream(), box, geometry, control, t, t_step, pf);

  logfile << "last restart file written." << std::endl;

  // ---------------------------------------------------
  // end simulation
  // ---------------------------------------------------

#ifdef PTM_VISUALIZE
  // keep the last configuration up until the user explicitly stops the 
  // program
  if ( mpi.node_id() == 0 && graphics_device == "OpenGL" ){
    std::cout << "press <ENTER> to exit...";
    std::cin.get();
  }
#endif 

  return 0;
}











