// $Id: main_eval.cc,v 1.6 2001/07/20 10:31:19 sts Exp $

// evaluate one or a sequence of 'save' files by calling the routine 
// 'ReadSaveFile' to read data into a 'box' and 'eval_data' with the 
// same arguments to evaluate them. The 'eval_data' routine is expected 
// to be a member of some class which can be used to do pre- and 
// postprocessing (output) of statistically analyzed data or similar 
// activities. This analysis is kept in a separate file and can thus 
// be exchanged 'easily'

// make sure we dont run parallel 
#if PT_PARALLEL > 0
#define PT_PARALLEL 0    // yields warning as well
#endif

#include <fstream>
#include <memory>

#include "PT/PTC_Para.h"
#include "PT/PTM_PartContLC.h"
#include "PT/PTM_ParPartCont.h"
#include "PT/PTSSimFile.h"
#include "PT/PT_Misc.h"

#include "ModParticle.H"  // typedef for double/int_vec 
#include "ModEvaluate.H"

#ifdef PTM_VISUALIZE
# include "XT/XTDisplay.h"
# include "XT/XTGLDev.h"
#endif 


#include "ModGeometry.H"  
#include "ModControl.H"  // defines ParticleBox
#include "ModIO.H"       // save/restart/measure 

#include "PT/PTDebug.h"


int main(int argc, char **argv)
{  
  // you can specify 'pfile=<filename>' in the command line, otherwise 
  // <progname>.cnf is expected to be present

  // movie=1/0 if you want the visulaization to be converted to a movie

  PTC_Para pf;          
  pf.standard_init(argc, const_cast<const char **>(argv)); 

  string strProject = pf.getd("project_name","");  
  // Project name specified?
  if (strProject==""){
    std::cerr << "No project_name specified in configuration file" << endl;
    return (-1);
  } 

  // print parameter settings
  pf.print_state(cout);       

  // process options
  // color scheme is particle_type, pressure, or kinetic_energy  
  string  color_scheme = pf.getd("color_scheme","particle_type");
  double  max_color    = pf.getd("max_color",10.);
  (void)  max_color;   // avoid warnings about unused variable   
  // for kinetic_energy or pressure scheme

  string  bitmap_type = pf.getd("bitmap_type",string("tiff"));
  bool    movie    = pf.getd("movie",false); // make tiffs for each snapshot
  (void)  movie;       // avoid warning about unused variable
  int     snapshot = pf.getd("snapshot",-1); // one tiff from picture with 
                                             //  sequence num 'snapshot'
  (void)  snapshot;
  string  pattern  = pf.getd("pattern",string("")); 
                                             // specify input files as pattern
  string  listfile = pf.getd("patternlist",string(""));
                       // or as a file containing an ordered list of filenames
                       // or patterns

  // if there is no pattern and no list, we create a pattern from 
  //  the project name
  if ( pattern == "" && listfile == "" ) 
    pattern  = strProject + "-save";

  // now if the creation mode for the save files requires distinct files 
  // for different configurations, we append a %d to be later replacd by the 
  // sequence 0, 1, 2 ...
  { 
    // a block to keep these names local
    string tmp = pf.getd("save","");
    stringstream in_str(tmp);
    double dummy; unsigned mode = 0;
    in_str >> dummy >> dummy >> mode;
    if ( in_str )   // input ok
      if ( mode & PTS_SimFileBase::MANY_FILES )
	pattern += "-%d";
  }

  // ------  material parameters from materials file:
  Material material(pf);
  WallFeature::material = material;

  // geometry / wall specification
  Geometry        geometry(pf);
  double_vec ll = geometry.ll;
  double_vec ur = geometry.ur;
  
#ifdef PTM_VISUALIZE

  std::auto_ptr<XTDevice> auto_w;
  std::string  graphics_device = pf.getd("graphics_device","OpenGL");
  if ( graphics_device == "OpenGL" ) {
    // ------ visualization stuff:
    //XTX11Dev w(500,500,argc,argv); // X11 2d graphics
    //XTPovDev w("out-movie");        // Pov Ray rendering
    std::auto_ptr<XTDevice> tmp(new XTGLDev(500,500,argc,argv)); 
    auto_w = tmp; 
  }
  else if ( graphics_device == "Null" ) {
    cout << "graphics_device is specified to be `Null',\n"
            "use `graphics_device=OpenGL' in the command line to override" 
	 << std::endl ;
  }

  XTDisplay  window(auto_w.get());                
  XTColormap map(XTColormap::rainbow);

  XTCrange   material_colors(0,10);
  // double     max_color = 0;  // for kientic_energy or pressure scheme

  if ( graphics_device != "Null" ) {
    window.set_colormap_mode(XTDevice::indexed);
    window.set_colormap(map);

    // user color range 
    window.set_color_range(material_colors);
    window.set_background_color(XTWhite);
    window.set_foreground_color(XTBlack);
    
    // coordinate range, perspective 
    double_vec center(0.5*(ur+ll)), span(ur-ll);
    // its  essential that the eye position is outside the coordiante range, 
    // even as we rotate
    double     max_span = PTmax(span[0],span[1],span[2]);
    XTPoint    eye_pos(center[0]+1.5*max_span,center[1]-3*max_span,center[2]);
    XTPoint    look_at(center[0],center[1],center[2]);
    XTPoint    up_dir(0,0,1); 
    XTObserver xo(eye_pos,look_at,up_dir);
    XTViewport vp;
    vp.set_pc(XTPoint(ll[0]-0.3*span[0],ll[1]-0.3*span[1],ll[2]-0.3*span[2]),
	      XTPoint(ur[0]+0.3*span[0],ur[1]+0.3*span[1],ur[2]+0.3*span[2]));
    vp.set_observer(xo);
    window  << vp; // set viewport
  }
#endif  

  // -- and create the empty particle container
  int_vec  bc = 0;
  double_vec extent(ur - ll);
  double cutoff = PTmin ( extent[0], extent[1], extent[2] ) / 10; 
  PTM_Parallel par_info(argc,argv);
  ParticleBox  *box1 = new ParticleBox(ll,ur,bc,cutoff,par_info);

  // a backup copy of the last configuration for tracing particles or 
  // for using the configuration to create a ball file or ...
  ParticleBox  *box2 = new ParticleBox(ll,ur,bc,cutoff,par_info);

  // select walls active in our coordinate range, cutoff is a measure of
  // of the 'range'/vicinity that needs to be taken into account, if 
  // in doubt specify something large compared to the particle radius
  geometry.select_features(cutoff,box1->local_ll(),box1->local_ur());
  cout << "box domain: " << ll << "/ " << ur << endl;;

  cout << "box.size() = " << box1->size() << " object(s)" << endl;
  cout << "geometry.size() = " << geometry.size() 
	  << " wall feature(s)" << endl;

  // ---------------------------------------------------
  // start evaluation sequence
  // ---------------------------------------------------

  double t=0.;          // simulation time  
  int    sequence_number = 0;  // current number of simulation frames read

  Evaluation  eval(pf,geometry,material);

  bool fast = false;   // fast movies

  // points to a list of file names/file name patterns if listfile exists
  PT_INFO("trying to open pattern list `" << listfile.c_str() << "'");
  istream *list    = new(std::nothrow) ifstream(listfile.c_str());
  istream *current = 0;
  if ( list && ! list->good() ) { 
    delete list; 
    list = 0; 
    PT_DETAILLED ( " ... failed " );
  }
  do {

    int  file_count = 0;
    do {
      // use the current pattern until we fail to open, then try to 
      // get a new one from the list. This also covers the case that we only 
      // got a pattern, but no list 
      // if the pattern contains %d, we will substitute a number starting at 
      // 0 and increasing until we fail to open the file
      string::size_type pos = pattern.find("%d");
      if ( pos != PT_STR_NPOS ){ // found it, replace it
	string fname = pattern;
	fname.replace( pos, 2, PTi2string(file_count++) );
	PT_INFO("trying to open " << fname.c_str());
	current = new(std::nothrow) ifstream(fname.c_str());
	
	if ( !current->good() ){ // no such file
	  PT_DETAILLED( "... failed" );
	  delete current; 
	  break;
	}
	else 
	  PT_DETAILLED( "... successful" );
	  
      }
      else {
	PT_INFO("trying to open " << pattern.c_str());
	current = new(std::nothrow) ifstream(pattern.c_str());
	if ( current->good() && file_count == 0 ){ 
	  // ok this is the first attempt
	  file_count++;
	  PT_DETAILLED( "... successful" );
	}
	else {
	  delete current; 
	  break;
	}

      }

      // read until eof, read returns 1 if there are problems or eof
      // empty box
      do { 
	
	// swap boxes
	ParticleBox *tmp = box1;
	box1 = box2; 
	box2 = tmp;

	box1->clear();
	if ( ! ReadSaveFile( *current, t, *box1, geometry, pf ) )
	  break;

	// show stuff 
	// evaulate
	cout << "seq_no= " << sequence_number << ", N/n = " 
	     << box1->global_size() << "/" << box1->size() << " ";

	// ---------------------
	// evaluate
	// ---------------------
	// this may add some output on the current line  

	eval.configuration(t,*box1,pf,geometry,material);

	cout << "\n";


	// ---------------------
	// end evaluation
	// ---------------------  
#ifdef PTM_VISUALIZE
    // clear window
	if ( graphics_device != "Null" ){
	  window.clear();
	  
	  if ( color_scheme == "kinetic_energy" ) { 
	    if ( ! pf.exists("max_color") ) {
	      if ( max_color == 0. ) max_color = eval.e_kin;
	      else max_color = 0.9 * max_color + 0.1 * eval.e_kin;
	    }	    
	    window << XTCrange(0.,5.*max_color); 

	    for ( ParticleBox::iterator i=box1->begin(); 
		  i != box1->all_end(); 
		  ++i ) {
	      XTPoint  center( (*i).x[0],(*i).x[1],(*i).x[2]);
	      window << XTColor( .5*material.getMass((*i).type)
				 * dot((*i).v,(*i).v) );
	      window << XTSphere( center, material.getRadius((*i).type) );
	    }
	    
	  } 
#if MOD_JANSEN_PRESSURE
	  else if ( color_scheme == "pressure" ) {

	    if ( ! pf.exists("max_color") ) { 
	      if ( max_color == 0. ) max_color = eval.jansen.max_pressure;
	      else max_color = 0.9 * max_color + 0.1 * eval.jansen.max_pressure;
	    }
	    window << XTCrange(0.,max_color); 

	    for ( ParticleBox::iterator i=box1->begin(); 
		  i != box1->all_end(); 
		  ++i ) {
	      XTPoint  center( (*i).x[0],(*i).x[1],(*i).x[2]);
	      window << XTColor( (*i).pressure );
	      window << XTSphere( center, material.getRadius((*i).type) );
	    }
	    
	  }
#endif	    
	  else {  // default: use material colors or external value
	    
	    window << XTCrange(0.,max_color); 

	    for ( ParticleBox::iterator i=box1->begin(); 
		  i != box1->all_end(); 
		  ++i ) {
	      XTPoint  center( (*i).x[0],(*i).x[1],(*i).x[2]);
	      window << XTColor( material.getColor((*i).type) );
	      window << XTSphere( center, material.getRadius((*i).type) );
	    }
	    
	  }
	  

	  // draw walls
	  window << material_colors;
	  geometry.update(t);
	  geometry.draw(window);
	  
	  // draw simulation box
	  window.set_foreground_color(XTBlack);
	  double_vec ur = box1->ur();
	  double_vec ll = box1->ll();
	  window.rectangle(XTRectangle(XTPoint(ll[0],ll[1],ll[2]),
				       XTPoint(ur[0],ur[1],ur[2])));
	  
	  // ... label the axes
	  window << XTText(XTPoint(ur[0],ll[1],ll[2]),"x");
	  window << XTText(XTPoint(ll[0],ur[1],ll[2]),"y");
	  window << XTText(XTPoint(ll[0],ll[1],ur[2]),"z");
	  
	  // show everything
	  window.display();

	}

	if ( snapshot == sequence_number ){
	  string fname = "snaphot-" + PTi2string(sequence_number) 
	                 + "." + bitmap_type;
          XTX11EvWin *ewin = dynamic_cast<XTX11EvWin *>(&(*(auto_w)));
	  if ( ewin )
	    ewin->save_file(fname.c_str());
	  else 
	    PT_WARNING( "graphics device does not support snapshots" );
	}  
	if ( movie ){
	  string fname = "movie-" + PTi2string(sequence_number) + "." + 
	    bitmap_type;
          XTX11EvWin *ewin = dynamic_cast<XTX11EvWin *>(&(*(auto_w)));
	  if ( ewin )
	    ewin->save_file(fname.c_str());
	  else 
	    PT_WARNING( "graphics device does not support snapshots" );
	}

#endif

	// interactive stuff - we'd like the messages before 
	// entering the loop
	bool done = false;
	while ( !done && !fast ) {
	  cout << "\n(c) continuous mode\n"
#if PTM_VISUALIZE
	          "(s) create snapshot.gif\n"
#endif
	          "(b) create `ballfile'\n"
                  "<ENTER> next frame\n"
                  " ";
	  char c = 'n';
	  do { 
	    char in[5] = {'n'};
	    cin.getline(in,5);
	    c = in[0];
	  } while ( isspace(c) && ! c == '\n' );
	  
	  if ( c == 'c' ) {
	    fast = true;
	    done = true;
	  }
	  else if ( c == 'b' ){ 
	    cout << "creating ball file `ballfile'";
	    cout.flush();
	    if ( ifstream("ballfile").good() ) // exists already
	      cout << " - cowardly refusing to overwrite\n";
	    else {
	      ofstream o("ballfile");
	      WriteBallFile(o,*box1);
	      cout << " - done\n";
	    }
	  }
#if PTM_VISUALIZE
	  else if ( c == 's' ) {
	    if ( dynamic_cast<XTX11EvWin *>(&(*(auto_w))) == 0 )
	      cout << "device does not support snapshots\n";
	    else { 
	      cout << "creating snapshot.gif";
	      cout.flush();
	      dynamic_cast<XTX11EvWin *>(&(*(auto_w)))->save_file("snapshot.gif");
	      cout << " - done.\n";
	    }
	  }
#endif
	  else 
	    done = true;

	} // while
      

	sequence_number++;
	
      }  while ( 1 ); // break on failure of ReadSaveFile
      delete current; 
      current = 0;

    } while ( 1 );  // break's when opening the files will get us out

    // file is exhausted, try to open the next one, list is 0 if 
    // the file could not be opened. 
    if ( list ) 
      *list >> pattern;
      
    // if the i/o has worked, then the condition for the while is true: 
  } while ( list && list->good() );


  bool done = false;
  while ( !done ) {
    cout << "(x) exit\n"
#if PTM_VISUALIZE
	    "(s) create snapshot.gif\n"
#endif
	    "(b) create `ballfile'\n"
            " ";
    char c;
    do { 
      char in[5];
      cin.getline(in,5);
      c = in[0];
    } while ( isspace(c) );
    
    if ( c == 'x' ) {
      // only exit...
      done = true;
    }
    else if ( c == 'b' ){ 
      cout << "creating ball file `ballfile'";
      cout.flush();
      if ( ifstream("ballfile").good() ) // exists already
        cout << " - cowardly refusing to overwrite\n";
      else {
	ofstream o("ballfile");
	WriteBallFile(o,*box2);
	cout << " - done\n";
      }
    }
#if PTM_VISUALIZE
    else if ( c == 's' ) 
      if ( dynamic_cast<XTX11EvWin *>(&(*(auto_w))) != 0 ) 
	dynamic_cast<XTX11EvWin *>(&(*(auto_w)))->save_file("snapshot.gif");
#endif
  
  } // while

  return 0;
}


